
import edward as ed
import inferpy as inf
import numpy as np
import tensorflow as tf
import random



K, d, N = 3, 4, 10
x_train = inf.models.Normal(loc=10, scale=100, dim=d).sample(N)

T=100

# MODEL
p2 = ed.models.Dirichlet(concentration=tf.ones(K)/K)
mu2 = ed.models.Normal(0.0, 1.0, sample_shape=[K, d])
sigma2 = ed.models.InverseGamma(concentration=1.0, rate=1.0,
                     sample_shape=[K, d])
z2 = ed.models.Categorical(logits=tf.log(p2) - tf.log(1.0 - p2), sample_shape=N)
z2 = ed.models.Categorical(probs=p2+1e-5, sample_shape=N)
x2 = ed.models.Normal(loc=tf.gather(mu2, z2), scale=tf.gather(sigma2, z2))



id = str(random.randint(0,100000))

def checknan(t):
    return tf.where(tf.is_nan(t), tf.zeros_like(t), t)
    #return t

# INFERENCE
qp_dist = ed.models.Dirichlet(concentration=checknan(tf.get_variable(
    "qp/params" + id,
    [K],
    initializer=tf.constant_initializer(1.0 / K))),allow_nan_stats=False)
qmu_dist = ed.models.Normal(loc=tf.get_variable("qmu/loc" + id,
                                       [K, d],
                                       initializer=tf.zeros_initializer()),
                       scale=tf.get_variable("qmu/scale" + id,
                                           [K, d],
                                           initializer=tf.zeros_initializer()),allow_nan_stats=False)
qsigma_dist = ed.models.InverseGamma(concentration=tf.get_variable("qsigma/conc"+id,
                                          [K,d],
                                          initializer=tf.ones_initializer()),
                                rate=tf.get_variable("qsigma/rate" + id,
                                                              sigma2.rate.shape,
                                                              initializer=tf.ones_initializer()),allow_nan_stats=False)
qz_dist = ed.models.Categorical(logits=
                                checknan(
                                           tf.get_variable("qz/params"+id,
                                      z2.logits.shape,
                                      #initializer=tf.constant([0.2, 0.2, 0.6]),
                                      initializer=tf.zeros_initializer(),
                                      dtype=tf.float32),
                                ),
    sample_shape=N, allow_nan_stats=False)


# en algún momento prob toma el valor nan y muestrea 3

#sess = ed.get_session()
#sess.run(tf.get_default_graph().get_operation_by_name("qz/params" + id))
#qz_dist.logits.eval()
# Tras inicializar la variable siempre muestrea el 3



latent_vars = {z2: qz_dist,
 p2: qp_dist,
 sigma2: qsigma_dist,
 mu2: qmu_dist}

#ed.KLpq(latent_vars=latent_vars, data={x2:x_train}).run()
ed.Laplace(latent_vars=latent_vars, data={x2:x_train}).run()


qz_dist.probs.eval()
qp_dist.concentration.eval()


[v.shape for v in latent_vars.keys()]
[v.shape for v in latent_vars.values()]




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


""" Module implementing text formatting operaionts """




import numpy as np
from inferpy.util.runtime import Runtime

def np_str(s):

    """ Shorten string representation of a numpy object

        Args:
            s: numpy object.

    """

    if not Runtime.compact_param_str:
        return str(s)

    old_ops = np.get_printoptions()
    np.set_printoptions(precision=3)
    np.set_printoptions(threshold=5)

    isnp = isinstance(s, np.ndarray)
    out = np.array_repr(s) if isnp else str(s)



    np.set_printoptions(precision=old_ops.get('precision'))
    np.set_printoptions(threshold=old_ops.get('threshold'))

    try:
        i = out.index(',')
    except ValueError:
        i = 0



    try:
        j = out.rindex('...')
    except ValueError:
        j = - 1


    if j>0:

        out = out[0:i] + "," + out[j:-1]

        try:
            k = out.rindex(',')
        except ValueError:
            k = len(out) - 1

        out = out[0:k]


    if isnp:
        out = out.replace("array(", "")


    return out



import inferpy as inf
import numpy as np
from inferpy.util.wrappers import def_ProbModel




# toy data generation
x_train = np.vstack([inf.models.Normal(loc=0, scale=1, dim=10).sample(300),
                     inf.models.Normal(loc=10, scale=1, dim=10).sample(700)])


####### define your custom parameterizable model #####


@def_ProbModel
def custom_gaussian_mixture(K,d,N):
    # prior distributions
    with inf.replicate(size=K):
        mu = inf.models.Normal(loc=0, scale=1,dim=d)
        sigma = inf.models.InverseGamma(concentration=1, rate=1, dim=d,)
    p = inf.models.Dirichlet(np.ones(K)/K)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Categorical(probs = p)
        x = inf.models.Normal(mu[z], sigma[z],observed=True, dim=d)


##### create instances of this model ####


m1 = custom_gaussian_mixture(K=3,d=10,N=1000)
m2 = custom_gaussian_mixture(K=2,d=10,N=1000)

#### compile each model ####

m1.compile()
m2.compile(infMethod="MCMC")

#### fit the data with each model #####

m1.fit({m1.observed_vars[0] : x_train})
m2.fit({m2.observed_vars[0] : x_train})




#############################################
#############################################

## you could also use the predefined models ##


from inferpy.models.predefined import *

m3 = pca(K=3,d=10,N=1000)

m3.compile()

m3.fit({m3.observed_vars[0] : x_train})









# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Module with the functionality related to inference methods.
"""

import inferpy as inf


INF_METHODS = ["KLpq", "KLqp", "Laplace", "ReparameterizationEntropyKLqp", "ReparameterizationKLKLqp", "ReparameterizationKLqp", "ScoreEntropyKLqp", "ScoreKLKLqp", "ScoreKLqp", "ScoreRBKLqp", "WakeSleep", "MetropolisHastings"]
""" List with all the alowed implemented inference methods """


INF_METHODS_ALIAS = {"Variational" : "KLqp", "MCMC":"MetropolisHastings"}
""" Aliases for some of the inference methods """


















import inferpy as inf
import tensorflow as tf





sess = inf.get_session()




# p(x|y)

with inf.ProbModel() as m:
    y =  inf.models.Categorical(probs=[0.4,0.6], name="y")
    x = inf.models.Categorical(probs=inf.case({y.equal(0): [0.0, 1.0],
                                               y.equal(1): [1.0, 0.0] }), name="x")

m.sample()





with inf.ProbModel() as m2:
    a = inf.models.Normal(0,1, name="a")
    b = inf.models.Categorical(probs=inf.case({a>0: [0.0, 1.0],
                                               a<=0: [1.0, 0.0]}), name="b")



m2.sample()




with inf.ProbModel() as m3:
    d =  inf.models.Categorical(probs=[0.4,0.6], name="d")
    c = inf.models.Normal(loc=inf.case({d.equal(0): 0.,
                                        d.equal(1): 100.}), scale=1., name="c")


m3.sample()

t = inf.models.Normal(0,5,dim=1)
y = inf.models.Beta([t,0.5], 1)
z =  inf.models.Categorical(probs=[0.4,0.6], name="z")


(x>0).loc


(x and x).sample()

(x>1 and x<10).sample()



type((t>0 and t<0)) # deterministic
type((x and x))     # categorical
type(x>0)           # deterministic

(t and t>0).loc     #normal

(t>0) == False

t





(x > [[0,0,0], [0,0,0]]).loc  # broadcasting is done,




y =  inf.models.Categorical(probs=[0.8,0.2], name="y", dim=2)

p = inf.case({y.equal(0): [0.5, 0.5], y.equal(1): [1.0, 0.0] })



p = tf.constant([0.1,0.8])


p = tf.gather([[0.5, 0.5], [0.1, 0.9]], y.dist)


y.equal([0,0]).loc


p = inf.case_states(y, {0: [0.5, 0.5], 1: [1.0, 0.0] })


#####

y =  inf.models.Categorical(probs=[0.5,0.5], name="y", dim=2)
p = inf.case_states(y, {(0,0): [1.0, 0.0, 0.0, 0.0], (0,1): [0.0, 1.0, 0.0, 0.0],
                        (1, 0): [0.0, 0.0, 1.0, 0.0], (1,1): [0.0, 0.0, 0.0, 1.0]} )


with inf.replicate(size=10):
    x = inf.models.Categorical(probs=p, name="x")


sess.run([y.dist,x.dist])

####

y =  inf.models.Categorical(probs=[0.5,0.5], name="y", dim=1)
z =  inf.models.Categorical(probs=[0.5,0.5], name="z", dim=1)


p = inf.case_states((y,z), {(0,0): [1.0, 0.0, 0.0, 0.0], (0,1): [0.0, 1.0, 0.0, 0.0],
                            (1, 0): [0.0, 0.0, 1.0, 0.0], (1,1): [0.0, 0.0, 0.0, 1.0]} )

with inf.replicate(size=10):
    x = inf.models.Categorical(probs=p, name="x")


sess.run([y.dist, z.dist, x.dist])

####

p = inf.case_states([y,z], {(0,0): [1.0, 0.0, 0.0, 0.0], (0,1): [0.0, 1.0, 0.0, 0.0],
                            (1, 0): [0.0, 0.0, 1.0, 0.0], (1,1): [0.0, 0.0, 0.0, 1.0]} )

with inf.replicate(size=10):
    x = inf.models.Categorical(probs=p, name="x")


sess.run([y.dist, z.dist, x.dist])






tf.shape(p)



x

sess = inf.get_session()

sess.run([y.dist, x.dist])

########
c = t


sess.run(tf.equal(c.base_object,True))


sess.run(tf.equal(z.base_object, 0))



sess.run(t>0)
z==0
sess.run(tf.equal((z==0 and t>0).base_object,True))

(t>0).base_object







t =  inf.models.Categorical(logits=[0,0], name="t")



x = inf.models.Categorical(probs=inf.case({t.equal(0): [0.5, 0.5],
                                           t.equal(1): [1.0, 0.0] }), name="x")




x = inf.models.Categorical(probs=[0.8, 0])

N = 100


s = x.sample(N,tf_run=False)


s = tf.reshape(x.sample(N,tf_run=False), [1,N])
s

import tensorflow as tf


x

tf.reduce_mean(tf.gather(tf.constant([10,100,500]), s)).eval()





# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


""" Module implementing the shared functionality across all the variable types  """


from inferpy.util import tf_run_wrapper
from inferpy.util import static_multishape
import inferpy as inf
import tensorflow as tf
import edward as ed
import numpy as np
import sys
import collections





class RandomVariable(object):
    """Base class for random variables.
    """


    __declared_vars = {}

    def __init__(self, base_object=None, observed=False):

        """ Constructor for the RandomVariable class

            Args:
                base_object: encapsulated Edward object (optional).
                observed (bool): specifies if the random variable is observed (True) or observed (False).



        """


        self.base_object = base_object
        self.__bind = None
        self.observed = observed
        self.copied_from = None

        if base_object != None:



            if  inf.ProbModel.is_active() and not self.is_generic_variable():
                inf.ProbModel.get_active_model().varlist.append(self)

            if inf.replicate.in_replicate():
                for r in inf.replicate.get_active_replicate():
                    r.varlist.append(self)

        RandomVariable.__declared_vars.update({id(self) : self})



    @property
    def dist(self):
        """Underlying Edward object"""

        if self.is_generic_variable():
            return None

        return self._base_object

    @property
    def base_object(self):
        """Underlying Tensorflow object"""

        return self._base_object



    @property
    def dim(self):
        """ Dimensionality of variable """
        return self.base_object.shape.as_list()[- (1 + len(self.event_shape)) ]

    @property
    def batches(self):
        """ Number of batches of the variable"""

        dist_shape = self.base_object.shape.as_list()

        if len(dist_shape) - len(self.event_shape) <= 1 :
            return 1
        return dist_shape[0]

    @property
    def shape(self):
        """ shape of the variable, i.e. (batches, dim)"""
        return self.base_object.shape.as_list()

    @property
    def event_shape(self):
        """ event_shape"""
        if self.is_generic_variable():
            return []

        return self.base_object.event_shape.as_list()


    @property
    def name(self):
        """ name of the variable"""
        return self.base_object.name[0:-1]


    @property
    def observed(self):
        """ boolean property that determines if a variable is observed or not """
        return self.__observed

    @observed.setter
    def observed(self,observed):
        """ modifies the boolean property that determines if a variable is observed or not """
        self.__observed=observed


    @property
    def bind(self):
        """  """
        return self.__bind

    @bind.setter
    def bind(self,bind):
        """  """

        if not isinstance(bind, RandomVariable):
            raise ValueError("object to bind is not RandomVariable")

        self.__bind=bind



    @dist.setter
    def dist(self, dist):
        """ Set the Underlying Edward object"""

        if isinstance(dist, ed.models.RandomVariable)==False:
            raise ValueError("Type of input distribution is not correct")

        self._base_object = dist

    @base_object.setter
    def base_object(self, tensor):
        """ Set the Underlying tensorflow object"""

        if isinstance(tensor, tf.Tensor)==False and \
                        isinstance(tensor, ed.RandomVariable) == False and tensor != None:
            raise ValueError("Type of input object is not correct")

        self._base_object = tensor


    @tf_run_wrapper
    def sample(self, size=1):
        """ Method for obaining a samples

        Args:
            size: scalar or matrix of integers indicating the shape of the matrix of samples.

        Return:
            Matrix of samples. Each element in the output matrix has the same shape than the variable.


        """

        if self.is_generic_variable():
            return self.base_object


        s = self.dist.sample(size)

        if self.dim > 1 and self.batches == 1:


            final_shape = (size, self.dim) if self.event_shape == [] else (size, self.dim, self.event_shape[0])

            s = tf.reshape(s, shape=final_shape)

        return s



    @tf_run_wrapper
    def prob(self, v):
        """ Method for computing the probability of a sample v (or a set of samples)"""
        return self.dist.prob(tf.cast(v, tf.float32))

    @tf_run_wrapper
    def log_prob(self, v):
        """ Method for computing the log probability of a sample v (or a set of samples)"""
        return self.dist.log_prob(tf.cast(v, tf.float32))

    @tf_run_wrapper
    def prod_prob(self, v):
        """ Method for computing the joint probability of a sample v (or a set of samples)"""
        return tf.reduce_prod(self.dist.prob(tf.cast(v, tf.float32)))

    @tf_run_wrapper
    def sum_log_prob(self, v):
        """ Method for computing the sum of the log probability of a sample v (or a set of samples)"""
        return tf.reduce_sum(self.dist.log_prob(tf.cast(v, tf.float32)))

    @tf_run_wrapper
    def mean(self, name='mean'):

        """ Method for obaining the mean of this random variable """

        if not self.is_generic_variable():
            return self.dist.mean(name)
        return None

    @tf_run_wrapper
    def variance(self, name='variance'):

        """ Method for obataining the variance of this random variable"""

        if not self.is_generic_variable():
            return self.dist.variance(name)
        return None

    @tf_run_wrapper
    def stddev(self, name='stddev'):

        """ Method for obataining the standard deviation of this random variable"""


        if not self.is_generic_variable():
            return self.dist.stddev(name)
        return None

    def is_generic_variable(self):
        """ Determines if this is a generic variable, i.e., an Edward variable
        is not encapsulated. """

        return isinstance(self._base_object, ed.RandomVariable) == False


    def get_replicate_list(self):

        """ Returns a list with all the replicate constructs that this variable belongs to.  """

        return [r  for r in inf.replicate.get_all_replicate() if self in r.varlist]


    def get_local_hidden(self):

        """ Returns a list with all the local hidden variables w.r.t. this one. Local hidden variables
        are those latent variables which are in the same replicate construct.  """

        var_rep = [r.varlist for r in self.get_replicate_list()]

        var_rep_id = [inf.models.RandomVariable.get_key_from_var(vr) for vr in var_rep]

        if len(var_rep) == 0:
            intersect_id = []
        elif len(var_rep) == 1:
            intersect_id = var_rep_id[0]
        else:

            for i in range(0, len(var_rep_id)):
                if i == 0:
                    intersect_id = var_rep_id[0]
                else:
                    intersect_id = np.intersect1d(intersect_id, var_rep_id[i])

        intersect_id = [x for x in intersect_id if inf.models.RandomVariable.get_key_from_var(self) != x]
        intersect = inf.models.RandomVariable.get_var_with_key(intersect_id)
        intersect = [x for x in intersect if not x.observed]

        return intersect

    def __repr__(self):

        """ Representation for this variable"""

        str = "<inferpy RandomVariable '%s' shape=%s dtype=%s>" % (
            self.name, self.shape, self.base_object.dtype.name)

        return str


    @staticmethod
    @static_multishape
    def get_var_with_key(key):
        return RandomVariable.__declared_vars.get(key)

    @staticmethod
    @static_multishape
    def get_key_from_var(var):

        from six import iteritems

        for (key, value) in iteritems(RandomVariable.__declared_vars):
            if value==var:
                return key
        return None


    def copy(self, swap_dict=None, observed=False):
        """ Build a new random variable with the same values.
        The underlying tensors or edward objects are copied as well.

        Args:
             swap_dict: random variables, variables, tensors, or operations to swap with.
             observed: determines if the new variable is observed or not.

        """

        new_var = getattr(sys.modules[self.__class__.__module__], type(self).__name__)()
        new_var.dist = ed.copy(self.dist, swap_dict)
        new_var.copied_from = self
        new_var.observed = False
        return new_var






# List of Python operators that we allow to override.
BINARY_OPERATORS = {
    # Binary.
    "__add__",
    "__radd__",
    "__sub__",
    "__rsub__",
    "__mul__",
    "__rmul__",
    "__div__",
    "__rdiv__",
    "__truediv__",
    "__rtruediv__",
    "__floordiv__",
    "__rfloordiv__",
    "__mod__",
    "__rmod__",
    "__lt__",
    "__le__",
    "__gt__",
    "__ge__",
    "__and__",
    "__rand__",
    "__or__",
    "__ror__",
    "__xor__",
    "__rxor__",
#    "__getitem__",
    "__pow__",
    "__rpow__",
    "__matmul__",
    "__rmatmul__"

}
UNARY_OPERATORS = {
    # Unary.
#    "__invert__",
    "__neg__",
    "__abs__"
}




def __add_operator(cls, name, unary=False):

    import inferpy.models.deterministic

    if unary==False:
        def operator(self, other):

            res = inferpy.models.Deterministic()

            if isinstance(other, RandomVariable):
                res.base_object = getattr(self.base_object, name)(other.base_object)
            else:
                res.base_object = getattr(self.base_object, name)(other)
            return res
    else:
        def operator(self):
            res = inferpy.models.Deterministic()
            res.base_object = getattr(self.base_object, name)()
            return res


    operator.__doc__ = "documentation for "+name
    operator.__name__ = name
    setattr(cls, operator.__name__, operator)


for x in BINARY_OPERATORS:
    __add_operator(RandomVariable,x)


for x in UNARY_OPERATORS:
    __add_operator(RandomVariable,x, unary=True)



def __add_equal_operator():

    import inferpy.models.deterministic

    name = "equal"
    cls = RandomVariable
    def operator(self, other):

        res = inferpy.models.Deterministic()

        op1 = self.base_object
        op2 = other.base_object if isinstance(other, RandomVariable) else other

        res.base_object = tf.equal(op1,op2)

        return res


    operator.__doc__ = "documentation for " + name
    operator.__name__ = name
    setattr(cls, operator.__name__, operator)


__add_equal_operator()





def __add_getitem_operator_old():

    import inferpy.models.deterministic

    name = "__getitem__"
    cls = RandomVariable
    def operator(self, index):

        res = inferpy.models.Deterministic()


        if isinstance(index, inf.models.RandomVariable):
            if index.batches == 1:

                if self.batches > 1:
                    res_dist = tf.reshape(tf.gather(self.base_object, index.base_object, axis=1),
                                          (self.batches,1))
                else:
                    res_dist = tf.reshape(tf.gather(self.base_object, index.base_object, axis=0), (1, 1))

            else:

                res_dist = tf.reshape(tf.stack([self[index[n,0]][n,0].base_object
                                                for n in range(0, index.batches)]), (index.batches, 1))


        else:

            if np.ndim(index) < 1:
                index = [index]

            if len(np.shape(index)) < 2:
                In = range(0, self.batches) if len(index) < 2 else [index[0]]
                Id = [index[-1]]

                if self.batches>1:
                    res_dist = tf.gather(tf.gather(self.base_object, In, axis=0), Id, axis=1)
                else:
                    res_dist = tf.reshape(tf.gather(self.base_object, Id, axis=0), (1,1))
            else:

                res_dist = tf.reshape(tf.stack([self[0,index[n][0]].base_object
                                                for n in range(0, np.shape(index)[0])]), (self.batches,1))


        res.base_object = res_dist

        return res



    operator.__doc__ = "documentation for " + name
    operator.__name__ = name
    setattr(cls, operator.__name__, operator)





def __add_getitem_operator():

    import inferpy.models.deterministic

    name = "__getitem__"
    cls = RandomVariable
    def operator(self, index):

        res = inferpy.models.Deterministic()

        res_tf = self.dist

        if not isinstance(index, collections.Iterable):
            index = [index]

        axis = 0
        ndim = len(self.shape)

        for i in index:
            if isinstance(i, inf.models.RandomVariable):
                res_tf = tf.gather(res_tf, i.base_object, axis=axis)
            elif isinstance(i, tf.Tensor):
                res_tf = tf.gather(res_tf, i, axis=axis)
            else:

                I = tuple([slice(None,None,None) for x in range(0,axis)] + [i])
                res_tf = res_tf[I]

            ndim_new = len(res_tf._shape_as_list())

            if ndim != ndim_new:
                ndim = ndim_new
            else:
                axis = axis+1
        res.base_object = tf.reshape(res_tf, inf.fix_shape(res_tf.shape))

        return res





    operator.__doc__ = "documentation for " + name
    operator.__name__ = name
    setattr(cls, operator.__name__, operator)



__add_getitem_operator()




"""
        if np.ndim(index)<1:
            index = [index]
        elif np.ndim(index)>1:
            raise ValueError("wrong index dimensions "+str(index))


        res_dist = self.base_object

        for i in index:
            res_dist = tf.gather(res_dist, i if not isinstance(i, inf.models.RandomVariable) else i.dist)

        res.base_object = res_dist

        return res

"""


import edward as ed
import inferpy as inf
from inferpy.models import Normal, Bernoulli, Categorical
import numpy as np

d, N =  10, 500

#number of classes
K = 3

# model definition
with inf.ProbModel() as m:

    #define the weights
    w0 = Normal(0,1, dim=K)

    with inf.replicate(size=K):
        w = Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        x = Normal(0, 1, observed=True, dim=d)
        y = Bernoulli(logits = w0 + inf.matmul(x, w, transpose_b=True), observed=True)



# toy data generation
x_train = Normal(loc=0, scale=1, dim=d).sample(N)
y_train = Bernoulli(probs=np.random.rand(K)).sample(N)
data = {x.name: x_train, y.name: y_train}


# compile and fit the model with training data
m.compile()
m.fit(data)

print(m.posterior([w, w0]))





import inferpy as inf
import edward as ed
import tensorflow as tf




from inferpy.models.params import *


d=4
N=10
K=3

p = [0.2, 0.7, 0.1]
p = inf.models.Dirichlet(np.ones(K))


inf.models.Normal(0,1, dim=3).shape  # [3]
inf.models.Normal(0,1, batches=10).shape  # [10,1]
inf.models.Normal(0,1, batches=[5,2]).shape  # [10,1]
inf.models.Normal(0,1, dim=3, batches=10).shape  # [10,3]
inf.models.Normal(0,1, dim=3, batches=[10,5]).shape  # [50,3]


with inf.replicate(size=2):
    print(inf.models.Normal(0, 1, dim=3, batches=[10, 5]).shape)  # [100,3]


inf.models.Normal(0,1, dim=[1,2]).shape  # error




import edward as ed
import inferpy as inf
import numpy as np
import tensorflow as tf



K, d, N, T = 3, 4, 1000, 5000


# toy data generation
x_train = np.vstack([inf.models.Normal(loc=0, scale=1, dim=d).sample(300),
                     inf.models.Normal(loc=10, scale=1, dim=d).sample(700)])

######## Inferpy ##########


# model definition
with inf.ProbModel() as m:

    # prior distributions
    with inf.replicate(size=K):
        mu = inf.models.Normal(loc=0, scale=1,dim=d)
        sigma = inf.models.InverseGamma(concentration=1, rate=1, dim=d,)
    p = inf.models.Dirichlet(np.ones(K)/K)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Categorical(probs = p)
        x = inf.models.Normal(mu[z], sigma[z],observed=True, dim=d)

# compile the model
m.compile(infMethod="MCMC")

# fit the model with training data
data = {x: x_train}
m.fit(data)

# print the posterior
print(m.posterior(mu))



######## Edward ##########



# model definition

# prior distributions
p = ed.models.Dirichlet(concentration=tf.ones(K)/K)
mu = ed.models.Normal(0.0, 1.0, sample_shape=[K, d])
sigma = ed.models.InverseGamma(concentration=1.0,rate=1.0, sample_shape=[K, d])

# define the generative model
z = ed.models.Categorical(logits=tf.log(p) - tf.log(1.0 - p), sample_shape=N)
x = ed.models.Normal(loc=tf.gather(mu, z), scale=tf.gather(sigma, z))

# compile the model
qp = ed.models.Empirical(params=tf.get_variable( "qp/params", [T, K],
                                                 initializer=tf.constant_initializer(1.0 / K)))
qmu = ed.models.Empirical(params= tf.get_variable("qmu/params", [T, K, d],
                                                  initializer= tf.zeros_initializer()))
qsigma = ed.models.Empirical( params= tf.get_variable("qsigma/params", [T, K, d],
                                                      initializer= tf.ones_initializer()))
qz = ed.models.Empirical( params= tf.get_variable("qz/params", [T, N],
                                                  initializer= tf.zeros_initializer(),
                                                  dtype=tf.int32))

gp = ed.models.Dirichlet(concentration=tf.ones(K))
gmu = ed.models.Normal(loc=tf.ones([K,d]), scale=tf.ones([K,d]))
gsigma = ed.models.InverseGamma(concentration= tf.ones([K,d]), rate=tf.ones([K,d]))
gz = ed.models.Categorical(logits=tf.zeros([N, K]))


# fit the model with training data
inference = ed.MetropolisHastings(
    latent_vars={p: qp, mu: qmu,
                 sigma: qsigma, z: qz},
    proposal_vars={p: gp, mu: gmu,
                   sigma: gsigma, z: gz},
    data={x: x_train})

inference.run()

# print the posterior
print(qmu.params.eval())




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""The Deterministic distribution """

import edward.models as base_models
import numpy as np
import tensorflow as tf

import inferpy.util
from inferpy.models.random_variable import *
from inferpy.util import tf_run_wrapper



class Deterministic(RandomVariable):

    """ Class implementing a Deterministic variable.

    This allows to encapsulate any Tensor object of Edward variable. Moreover, variables
    of this type can be intilially empty and later initialized. When operating with InferPy
    random variables, the result is always a deterministic variable.

    An example of use:

    .. literalinclude:: ../../examples/deterministic_variable.py






    """

    def __init__(self, loc=None, dim=None, observed=False, name="Determ", ):

        """Constructor for the Deterministic distribution


        """

        if loc != None:

            self.__check_params(loc, dim)   # loc and tensor cannot be None at the same time

            param_dim = 1
            if dim != None: param_dim = dim

            # shape = (batches, dimension)
            self_shape = (inf.replicate.get_total_size(), np.max([self.__get_total_dimension(loc), param_dim]))

            loc_rep = self.__reshape_param(loc, self_shape)

            # build the distribution
            super(Deterministic, self).__init__(base_models.Deterministic(loc=loc_rep, name=name), observed=observed)

        else:

            super(Deterministic, self).__init__(observed=observed)




    @property
    @tf_run_wrapper
    def loc(self):
        """Distribution parameter for the mean."""

        if self.dist != None:
            return self.dist.loc

        return self.base_object



    def __check_params(self, loc, dim):
        """private method that checks the consistency of the input parameters"""


        # loc  cannot be multidimensional arrays (by now)
        if np.ndim(loc) > 1:
            raise ValueError("loccannot be a  multidimensional arrays")


        dim_loc = self.__get_total_dimension(loc)


        # loc can be a scalar or a vector of length dim
        if dim != None and dim_loc > 1 and dim != dim_loc:
            raise ValueError("loc length is not consistent with value in dim")




    def __reshape_param(self,param, self_shape):

        N = self_shape[0]
        D = self_shape[1]


        # get a D*N unidimensional vector

        if np.shape(param) in [(), (1,)] or\
                (isinstance(param, inferpy.models.RandomVariable) and param.dim==1):
            param_vect = np.repeat(param, D * N).tolist()
        else:
            param_vect = np.tile(param, N).tolist()


        if np.all(list(map(lambda x: np.isscalar(x), param_vect))):            # only numerical values

            # reshape the list
            if N > 1:
                param_np_mat = np.reshape(np.stack(param_vect), (N, -1))
            else:
                param_np_mat = np.reshape(np.stack(param_vect), (D,))

            #transform in tf
            param_tf_mat = tf.constant(param_np_mat, dtype="float32")

        else:                                                           # with a tensor

            # transform the numerical values into tensors
            for i in range(0, len(param_vect)):
                if np.isscalar(param_vect[i]):
                    param_vect[i] = [tf.constant(param_vect[i], dtype="float32")]
                elif isinstance(param_vect[i], inferpy.models.RandomVariable):
                    param_vect[i] = param_vect[i].base_object

            # reshape the list
            if N>1:
                param_tf_mat = tf.reshape(tf.stack(param_vect), (N, -1))
            else:
                if D>1:
                    param_tf_mat = tf.reshape(tf.stack(param_vect), (D,))
                else:
                    param_tf_mat = param_vect[0]


        return param_tf_mat


    def __repr__(self):
        return "<inferpy Deterministic '%s', shape=%s>" % (
            self.name, self.shape)



    def __get_total_dimension(self, x):

        D = 0

        if np.ndim(x) == 0:
            x = [x]

        for xi in x:
            if np.isscalar(xi):
                D = D + 1
            elif isinstance(xi, inferpy.models.RandomVariable):
                D = D + xi.dim
            elif isinstance(xi, tf.Tensor):
                D = D + xi.get_shape().as_list()[-1]

            else:
                raise ValueError("Wrong input type")


        return D




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Module with useful wrappers used for the development of InferPy.
"""



from functools import wraps
import inferpy as inf
import numpy as np
from six import iteritems

import pandas as pd



def tf_run_wrapper(f):

    """ When setted to a function f, this wrappers replaces the output tensor of f by its evaluation
    in the default tensorflow session. In doing so, the API user will only work with standard Python
    types.
     """


    @wraps(f)
    def wrapper(*args, **kwargs):

        if "tf_run" in kwargs:
            tf_run = kwargs.pop("tf_run")
        else:
            tf_run = inf.util.Runtime.tf_run_default


        if tf_run:

            # transforms in a list
            output_tf = f(*args, **kwargs)
            if type(output_tf).__module__ == np.__name__ or type(output_tf).__name__ == list.__name__:
                output_tf_vect = output_tf
            elif type(output_tf).__name__ == dict.__name__:
                output_tf_vect = list(output_tf.values())
            else:
                output_tf_vect = [output_tf]

            # evaluation
            output_eval_vect = inf.util.Runtime.tf_sess.run(output_tf_vect)


            # transforms in original type
            if type(output_tf).__module__ == np.__name__ or type(output_tf).__name__ == list.__name__:
                output_eval = output_eval_vect
            elif type(output_tf).__name__ == dict.__name__:
                output_eval = {}
                i = 0
                for k, v in iteritems(output_tf):
                    output_eval.update({k: output_eval_vect[i]})
                    i = i+1
            else:
                output_eval = output_eval_vect[0]

            return output_eval
        return f(*args, **kwargs)
    return wrapper




def multishape(f):

    """ This wrapper allows to apply a function with simple parameters, over multidimensional ones. """

    @wraps(f)
    def wrapper(*args, **kwargs):

        first_arg = 1

        if np.ndim(args[first_arg]) == 0:            # single element
            return f(*args, **kwargs)
        elif np.ndim(args[first_arg]) == 1:        # unidimensional vector
            output = []
            for i in args[1]:

                if first_arg == 1:
                    output.append(f(args[0], i, **kwargs))
                else:
                    output.append(f(i, **kwargs))
            return output
        else:
            raise ValueError("@multishape wrapper can only be applied to single elements or to 1-dimension vectors")


    return wrapper



def static_multishape(f):
    """ This wrapper allows to apply a function with simple parameters, over multidimensional ones. """


    @wraps(f)
    def wrapper(*args, **kwargs):

        first_arg = 0

        if np.ndim(args[first_arg]) == 0:            # single element
            return f(*args, **kwargs)
        elif np.ndim(args[first_arg]) == 1:        # unidimensional vector
            output = []
            for i in args[first_arg]:

                if first_arg == 1:
                    output.append(f(args[0], i, **kwargs))
                else:
                    output.append(f(i, **kwargs))
            return output
        else:
            raise ValueError("@multishape wrapper can only be applied to single elements or to 1-dimension vectors")


    return wrapper





def singleton(class_):
    """ wrapper that allows to define a singleton class """

    class class_w(class_):
        _instance = None
        def __new__(class_, *args, **kwargs):
            if class_w._instance is None:
                class_w._instance = super(class_w,
                                    class_).__new__(class_,
                                                    *args,
                                                    **kwargs)
                class_w._instance._sealed = False
            return class_w._instance
        def __init__(self, *args, **kwargs):
            if self._sealed:
                return
            super(class_w, self).__init__(*args, **kwargs)
            self._sealed = True
    class_w.__name__ = class_.__name__
    return class_w




def input_model_data(f):

    """ wrapper that transforms, if required, a dataset object, making it suitable for InferPy inference
    process.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):

        self = args[0]
        data = args[1]

        if isinstance(data, pd.DataFrame):
            newdata =  {}

            for k, v in iteritems(data.to_dict(orient="list")):
                if self.get_var(k) != None:

                    newdata.update({k : np.reshape(v, (np.size(v),1))})


        elif isinstance(data, dict):
            newdata = data
        else:
            raise ValueError("Wrong input data type: it should be a pandas dataframe or a dictionary")


        return f(self,newdata)

    return wrapper




def def_ProbModel(f):

    """ wrapper for defining custom parametrizable models in functions
    """

    @wraps(f)
    def wrapper(*args, **kwargs):

        with inf.ProbModel() as m:
            f(*args, **kwargs)
        return m

    return wrapper





import edward as ed
import inferpy as inf
from inferpy.models import Normal
import numpy as np

d, N =  5, 20000

# model definition
with inf.ProbModel() as m:

    #define the weights
    w0 = Normal(0,1)
    w = Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        x = Normal(0, 1, observed=True, dim=d)
        y = Normal(w0 + inf.dot(x,w), 1.0, observed=True)


# toy data generation
x_train = inf.models.Normal(loc=10, scale=5, dim=d).sample(N)
y_train = np.matmul(x_train, np.array([10,10,0.1,0.5,2]).reshape((d,1))) \
          + inf.models.Normal(loc=0, scale=5, dim=1).sample(N)


data = {x.name: x_train, y.name: y_train}


# compile and fit the model with training data
m.compile()
m.fit(data)

print(m.posterior([w, w0]))





import inferpy as inf
import tensorflow.contrib.distributions as tfd
import edward as ed
import tensorflow as tf

sess = inf.get_session()

with inf.replicate(size=10):
    x = inf.models.MultivariateNormalDiag(
        loc=[1., -1],
        scale_diag=[1, 2.], dim=3
    )


x.shape         #  [10, 3, 2]
x.batches       #  10
x.dim           #  3
x.event_shape   # [2]




# -*- coding: utf-8 -*-
#

from inferpy.models.random_variable import *
#from inferpy.models.normal import *
from inferpy.models.deterministic import *
from inferpy.models.factory import *


from inferpy.models import predefined


# IMPLEMENTED RANDOM VARIABLES



ALLOWED_VARS = sorted(list(np.unique([cls.__name__ for cls in RandomVariable.__subclasses__()])))







import numpy as np
import tensorflow as tf
import inferpy as inf

from abc import abstractproperty, abstractmethod
import edward as ed
import collections



class Param(object):

    def __init__(self, name, value, is_simple=True, duplicate=None):
        self.name = name
        self.is_simple = True if is_simple in [None, True] else False

        elem_ndim = 0 if self.is_simple else 1


        self.p_value = new_ParamValue(value, elem_ndim)

        if duplicate != None:
            self.p_value = self.p_value.duplicate(duplicate)


        self.__input_value = value



    def __repr__(self):

        simple_str = "simple " if self.is_simple else "complex"

        return "<"+simple_str+" param "+self.name+"="+self.p_value.__repr__()+\
               ", ndim="+str(self.ndim)+", td="+str(self.total_dim)+", dim="+str(self.dim)+", bt="+str(self.batches)+">"

    @property
    def nd_range(self):
        if self.is_simple:
            return [0,1]
        return [1,2]

    @property
    def dim_elem(self):
        if self.is_simple:
            n =  1
        elif self.ndim==1:
            n = self.total_dim
        elif isinstance(self.p_value.value, tf.Tensor):
            n = self.p_value.value.get_shape().as_list()[-1]
        elif isinstance(self.p_value.value, collections.Iterable):
            n = self.p_value.value[0].total_dim
        elif isinstance(self.p_value.value, inf.models.RandomVariable):
            n = self.p_value.value.event_shape
            n = n[0] if len(n)>0 else 1
        else:
            n = 1


        return n


    @property
    def ndim(self):
        return self.p_value.ndim


    @property
    def total_dim(self):
        return self.p_value.total_dim

    @property
    def dim(self):
        return int(self.total_dim / self.dim_elem)

    @property
    def batches(self):
        return int(self.p_value.batches)


    @property
    def input_value(self):
        return self.__input_value





def new_ParamValue(value, elem_ndim=0):

    if type(value) in [type([]), type(np.array([]))]:
        p = ParamValueArray(value, elem_ndim)
    elif np.isscalar(value):
        p = ParamValueScalar(value)
    elif isinstance(value, tf.Tensor):
        p = ParamValueTensor(value)
    elif isinstance(value, inf.models.RandomVariable):
        p = ParamValueInfVar(value)
    elif isinstance(value, ed.models.RandomVariable):
        p = ParamValueTensor(value)

    else:
        raise ValueError("Wrong parameter value")

    return p

class ParamValue(object):

    @abstractproperty
    def ndim(self):
        pass

    @abstractproperty
    def total_dim(self):
        pass


    @abstractproperty
    def batches(self):
        pass

    def __repeat(self, k):
        v = [p for p in np.repeat(self, k)]
        ret = ParamValueArray([], elem_ndim=None)
        ret.value = v
        return ret

    def repeat(self, kd=1, kb=1):
        ret = self

        if kb == 1 and kd == 1:
            return ret




        if kd > 1 or kb > 1 and self.ndim==0: ret = ret.__repeat(kd)
        if kb > 1: ret = ret.__repeat(kb)

        return ret


    def __repr__(self):
        return str(self.value)


    def get_std_value(self):
        return self.value

    def get_param_tensor(self):
        return self.tensor







class ParamValueScalar(ParamValue):

    def __init__(self,value):
        self.value = value

    @abstractproperty
    def ndim(self):
        return 0


    @abstractproperty
    def total_dim(self):
        return 1


    @abstractproperty
    def batches(self):
        return 1

    def all_scalar(self):
        return True

    @property
    def tensor(self):
        return tf.constant([self.value], dtype="float32")

    def get_param_tensor(self):
        return tf.constant([self.value], dtype="float32")

    def duplicate(self, op="equal"):

        if op=="equal":
            new_value = self.value
        elif op=="negative":
            new_value = -self.value
        elif op == "prob_complement":
            new_value = 1-self.value


        return new_ParamValue([self.value, new_value], 0)


class ParamValueArray(ParamValue):

    def __init__(self, value, elem_ndim=0):
        self.value = []
        for v in value:
            child_p = new_ParamValue(v)

        #    if child_p.ndim != elem_ndim and elem_ndim != None:
        #        raise ValueError("Array Parameter with wrong number of dimension")

            self.value.append(child_p)

    @abstractproperty
    def ndim(self):
        return np.max([v.ndim+1 for v in self.value])

    @abstractproperty
    def total_dim(self):
        return np.sum([v.total_dim for v in self.value])


    @abstractproperty
    def batches(self):
        return self.value[0].total_dim


    def all_scalar(self):
        return np.all([v.all_scalar() for v in self.value])

    def get_std_value(self):
        return [v.get_std_value() for v in self.value]


    @property
    def tf_array(self):
        return np.array([v.tf_array if isinstance(v, ParamValueArray) else v.tensor for v in self.value])

    @property
    def tensor(self):
        if self.all_scalar():
            v = np.array(self.get_std_value())
            return tf.constant(v, dtype="float32")
        else:

            tf_array = self.tf_array
            shape = tf_array.shape
            tf_vect = list(tf_array.flatten())

            if shape == (1,):
                return tf_vect[0]

            return tf.stack(tf_vect)

         #   m = np.prod([x for x in tf_array.shape])
         #   if  m != self.total_dim:
         #       shape = tuple([x for x in tf_array.shape] + [self.total_dim / m])

        #    event_s = self.tf_array[0].event_shape.as_list() if hasattr(self.tf_array[0], "event_shape") else []
        #    if len(event_s)>0: shape = [x for x in shape] + event_s



        #    return tf.reshape(tf.stack(tf_vect), shape)




class ParamValueTensor(ParamValue):

    def __init__(self, value):
        if value.shape == ():
            value = tf.reshape(value, shape=(1,))

        self.value = value

    @abstractproperty
    def ndim(self):
        shape = self.value.shape.as_list()
        return 0 if shape in ([], [1])  else  len(shape)


    @abstractproperty
    def total_dim(self):
        return 1 if len(self.value.shape.as_list()) == 0 else self.value.shape.as_list()[-1]

    @abstractproperty
    def batches(self):
        return 1 if len(self.value.shape.as_list()) < 2 else self.value.shape.as_list()[0]

    def all_scalar(self):
        return False

    @property
    def tensor(self):
        return self.value

    def get_param_tensor(self):

        v = self.value

        if self.value.shape == ():
            v = tf.reshape(v, shape=(1,1))

        return v

    def duplicate(self, op="equal"):

        if op == "equal":
            new_value = self.value
        elif op == "negative":
            new_value = -self.value
        elif op == "prob_complement":
            new_value = 1 - self.value

        return new_ParamValue([self.value, new_value], 0)


class ParamValueInfVar(ParamValue):

    def __init__(self, value):
        self.value = value

    @abstractproperty
    def ndim(self):
        shape = self.value.shape
        return 0 if shape in ([], [1])  else  len(shape)

    @abstractproperty
    def total_dim(self):
        e = np.sum(self.value.event_shape)
        d = self.value.dim * e if e >0 else self.value.dim
        return d

    @abstractproperty
    def batches(self):
        return self.value.batches

    def all_scalar(self):
        return False

    @property
    def tensor(self):
        return self.value.base_object



class ParamValueEdVar(ParamValue):

    def __init__(self, value):
        if value.shape == ():
            value = tf.reshape(value, shape=(1,))


    @abstractproperty
    def ndim(self):
        shape = self.value.shape
        return 0 if shape in ([], [1])  else  len(shape)

    @abstractproperty
    def total_dim(self):
        return 1 if len(self.value.shape.as_list()) == 0 else self.value.shape.as_list()[-1]

    @abstractproperty
    def batches(self):
        return 1 if len(self.value.shape.as_list()) < 2 else self.value.shape.as_list()[0]

    def all_scalar(self):
        return False

    @property
    def tensor(self):
        return self.value

    def get_param_tensor(self):

        v = self.value

        if self.value.shape == ():
            v = tf.reshape(v, shape=(1,1))

        return v





class ParamList(object):

    def __init__(self, params, args_list=[], kwargs_dict={}, is_simple={}, param_dim=None):

        plist=[]

        for p_name in params:
            if len(args_list) > 0:
                if p_name in kwargs_dict:
                    raise ValueError("Wrong positional or keyword argument")

                plist.append(Param(name=p_name, value = args_list[0], is_simple=is_simple.get(p_name)))
                args_list = args_list[1:]
            else:
                if p_name in kwargs_dict:
                    plist.append(Param(name=p_name, value = kwargs_dict.get(p_name), is_simple=is_simple.get(p_name)))


        self.plist=plist
        self.param_dim = param_dim


    def __repr__(self):
        str = ""
        for p in self.plist:
            str =str+" "+p.__repr__()+",\n"

        return "["+str[1:-2]+"]"


    def is_empty(self):
        return len(self.plist)==0


    def check_params(self):
        self.__check_ndim()
        self.__check_batches()
        self.__check_dim()


    def __check_ndim(self):
        for p in self.plist:

            n = p.ndim
            low = p.nd_range[0]
            up = p.nd_range[1]

            if p.batches>1 and p.batches==inf.replicate.get_total_size():
                low = low+1
                up = up+1

            if n < low or n > up:
                raise ValueError(
                    "Wrong dimension of "+p.name+" which is " + str(n) + " but should be in interval [" + str(low) + "," + str(
                        up) + "]")


    def __check_batches(self):

        N = self.final_shape[0]
        param_batches = np.unique([p.batches for p in self.plist])



        if len([x for x in param_batches if x not in [1,N]]) > 0:
            raise ValueError("Error: the number of batches for all the parameters must be 1 or "+str(N))



    def __check_dim(self):
        D = self.final_shape[1]
        param_dim = np.unique([p.dim for p in self.plist])

        if len([x for x in param_dim if x not in [1,D]]) > 0:
            raise ValueError("Inconsistent parameter dimensions")


    @property
    def final_shape(self):
        N = inf.replicate.get_total_size()
        D = np.max([v.dim for v in self.plist])

        if self.param_dim != None:
            D = np.max([D, self.param_dim])




        return [N,D]


    @property
    def final_shape2(self):
        N = inf.replicate.get_total_size()
        D = np.max([v.dim for v in self.plist])

        if self.param_dim != None:

            nd = np.ndim(self.param_dim)

            if nd == 0:
                D = np.max([D, self.param_dim])
            elif nd == 1 and len(self.param_dim)==1:
                D = np.max([D, self.param_dim[0]])
            elif nd == 1 and len(self.param_dim)==2 and N ==1:
                N = self.param_dim[0]
                D = np.max([D, self.param_dim[1]])
            else:
                raise ValueError("ERROR: wrong shape of dim parameter")

        return [N,D]


    def get_reshaped_param_dict(self):
        d = {}
        final_shape = self.final_shape

        D = final_shape[1]
        N = final_shape[0]

        for p in self.plist:

            v = p.p_value.repeat(D / p.dim, N / p.batches).get_param_tensor()

            # complex params should have at least 2 dimensions
            if not p.is_simple and len(v.shape.as_list())<2:
                v = tf.stack([v])

            if not p.is_simple:
                n = p.dim_elem
                final_shape = [N, D,n] if N>1 else [D,n]
            else:
                final_shape = [N,D] if N>1 else [D]

            v = tf.reshape(v,final_shape)



            d.update({p.name : v})

        return d






import inferpy as inf
import numpy as np

K = 4   # extreme points
N = 1   # num vars
M = 3   # states per var


with inf.ProbModel() as m:

    with inf.replicate(K):
        px = inf.models.Normal(np.ones(M) / M, scale=1)

    with inf.replicate(K*N):

        tx = inf.models.Categorical(logits=np.zeros(K), observed=True)
        x = inf.models.Categorical(probs = px[tx])



train_tx = np.arange(0,K,1).reshape((K,1))



m.compile(infMethod='MCMC')
m.fit({tx : train_tx})


tx.sample()

tx.shape

px[1]




import inferpy as inf
import numpy as np
from inferpy.util.wrappers import def_ProbModel



@def_ProbModel
def gaussian_mixture(K,d,N):
    # prior distributions
    with inf.replicate(size=K):
        mu = inf.models.Normal(loc=0, scale=1,dim=d)
        sigma = inf.models.InverseGamma(concentration=1, rate=1, dim=d,)
    p = inf.models.Dirichlet(np.ones(K)/K)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Categorical(probs = p)
        x = inf.models.Normal(mu[z], sigma[z],observed=True, dim=d)





@def_ProbModel
def linear_regression(d,N):

    #define the weights
    w0 = inf.models.Normal(0,1)
    w = inf.models.Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        x = inf.models.Normal(0, 1, observed=True, dim=d)
        y = inf.models.Normal(w0 + inf.dot(x,w), 1.0, observed=True)





@def_ProbModel
def log_regression(d,N):

    #define the weights
    w0 = inf.models.Normal(0,1)
    w = inf.models.Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        x = inf.models.Normal(0, 1, observed=True, dim=d)
        y = inf.models.Bernoulli(logits=w0+inf.dot(x, w), observed=True)





@def_ProbModel
def log_regression(K,d,N):

    #define the weights
    w0 = inf.models.Normal(0,1, dim=K)

    with inf.replicate(size=K):
        w = inf.models.Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        x = inf.models.Normal(0, 1, observed=True, dim=d)
        y = inf.models.Bernoulli(logits = w0 + inf.matmul(x, w, transpose_b=True), observed=True)



@def_ProbModel
def pca(K,d,N):
    #define the weights
    with inf.replicate(size=K):
        w = inf.models.Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Normal(0, 1, dim=K)
        x = inf.models.Normal(inf.matmul(z,w),
                               1.0, observed=True, dim=d)



@def_ProbModel
def pca_with_ard_prior(K,d,N):
    #define the weights
    with inf.replicate(size=K):
        w = inf.models.Normal(0, 1, dim=d)

    sigma = inf.models.InverseGamma(1.0,1.0)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Normal(0, 1, dim=K)
        x = inf.models.Normal(inf.matmul(z,w),
                   sigma, observed=True, dim=d)


import inferpy as inf
import edward as ed
import tensorflow as tf


# empty initialized variable
x = inf.models.Deterministic()
x.dist = ed.models.Normal(0.,1.)


# deterministic variable with returning a tensor

x = inf.models.Deterministic()
x.base_object = tf.constant(1.)


# deterministic variable returning an InferPy variable

a = inf.models.Normal(0,1)
b = inf.models.Normal(2,2)

x = a + b

type(x)  # <class 'inferpy.models.deterministic.Deterministic'>





import inferpy as inf
import edward as ed
import tensorflow as tf


theta = inf.models.Normal(0,1)
with inf.replicate(size=10):
    x = inf.models.Normal(0,1, dim=4)


theta = ed.mo


# batches autocomplete!!
x = inf.models.Normal(scale = [theta, 2*theta, 3, tf.constant(1, dtype="float32")], loc= [1,1] , batches=10)

x.scale




# tensorboard --logdir= ./log
with tf.Session() as sess:
    writer = tf.summary.FileWriter("./log/", sess.graph)
    sess.run(tf.global_variables_initializer())



import edward as ed
import inferpy as inf
import numpy as np
import tensorflow as tf



K, d, N, T = 3, 4, 1000, 5000


# toy data generation
x_train = np.vstack([inf.models.Normal(loc=0, scale=1, dim=d).sample(300),
                     inf.models.Normal(loc=10, scale=1, dim=d).sample(700)])

######## Inferpy ##########


# model definition
with inf.ProbModel() as m:

    # prior distributions
    with inf.replicate(size=K):
        mu = inf.models.Normal(loc=0, scale=1, dim=d)
        sigma = inf.models.InverseGamma(concentration=1, rate=1, dim=d,)
    p = inf.models.Dirichlet(np.ones(K)/K)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Categorical(probs = p)
        x = inf.models.Normal(mu[z], sigma[z],observed=True, dim=d)

# compile and fit the model with training data
data = {x: x_train}
m.compile(infMethod="MCMC")
m.fit(data)

# print the posterior
print(m.posterior(mu))





import edward as ed
import inferpy as inf
import numpy as np
import tensorflow as tf
import random



K, d, N, T = 3, 4, 1000, 5000
x_train = inf.models.Normal(loc=10, scale=100, dim=d).sample(N)


# MODEL
p = ed.models.Dirichlet(concentration=tf.ones(K)/K)
mu = ed.models.Normal(0.0, 1.0, sample_shape=[K, d])
sigma = ed.models.InverseGamma(concentration=1.0, rate=1.0,
                     sample_shape=[K, d])
z = ed.models.Categorical(logits=tf.log(p) - tf.log(1.0 - p), sample_shape=N)
x = ed.models.Normal(loc=tf.gather(mu, z), scale=tf.gather(sigma, z))


# INFERENCE
qp = ed.models.Empirical(params=tf.get_variable(
    "qp/params",
    [T, K],
    initializer=tf.constant_initializer(1.0 / K)))
qmu = ed.models.Empirical(params=tf.get_variable("qmu/params",
                                       [T, K, d],
                                       initializer=tf.zeros_initializer()))
qsigma = ed.models.Empirical(params=tf.get_variable("qsigma/params",
                                          [T, K, d],
                                          initializer=tf.ones_initializer()))
qz = ed.models.Empirical(params=tf.get_variable("qz/params",
                                      [T, N],
                                      initializer=tf.zeros_initializer(),
                                      dtype=tf.int32))


gp = ed.models.Dirichlet(concentration=tf.ones(K))
gmu = ed.models.Normal(loc=tf.ones([K,d]),
             scale=tf.ones([K,d]))
gsigma = ed.models.InverseGamma(concentration=tf.ones([K,d]),
                      rate=tf.ones([K,d]))
gz = ed.models.Categorical(logits=tf.zeros([N, K]))



inference = ed.MetropolisHastings(
    latent_vars={p: qp, mu: qmu, sigma: qsigma, z: qz},
    proposal_vars={p: gp, mu: gmu, sigma: gsigma, z: gz},
    data={x: x_train})

inference.run()


print(qmu.params.eval())


import inferpy as inf
import pandas as pd
import edward as ed
import numpy as np
import tensorflow.contrib.distributions as tfd

N, K, d = 100, 5, 10



# variables

# option 1:


beta = inf.models.MultivariateNormalDiag(loc=[0,0,0], scale_diag=[1,1,1], dim=3)



with inf.replicate(size = N):
    z =  inf.models.Normal(loc=0, scale=1)   # M,H
    x = inf.models.Bernoulli(z*beta, 1, dim=d, observed=True) # M,d, ... k

# option 2:


beta = inf.models.Normal(loc=0, scale=1)

with inf.replicate(size = N):
    z =  inf.models.Normal(loc=0, scale=1)
    x = inf.models.Binomial(total_count=1, logits= z*beta, dim=d, observed=True)




# model definition
m = inf.ProbModel(varlist=[beta,z,x])
m.compile()


# infer the parameters from data
data = pd.read_csv("inferpy/datasets/test.csv")
m.fit(data)


print(m.posterior(beta).loc)


"""
>>> x.shape
[1000, 10]
>>> x.shape
[1000, 10]
>>> x.dim
10
>>> x.batches
1000
>>> beta.loc
array([0.], dtype=float32)
>>> beta.scale
array([1.], dtype=float32)

"""





import inferpy as inf
from inferpy.models import Normal

N=200
M=50
K=5

# Shape [M,K]
with inf.replicate(size=K):
    gamma = Normal(0,1, dim=M)

# Shape [N,K]
with inf.replicate(size=N):
    w = Normal(0,1, dim=K)

# x_mn has shape [N,K] x [K,M] = [N,M]

with inf.replicate(size=N):
    x = Normal(inf.matmul(w,gamma), 1, observed = True)

m = inf.ProbModel([w,gamma,x])

data = m.sample(size=N)

log_prob = m.log_prob(data)

m.compile(infMethod = 'KLqp')

m.fit(data)

print(m.posterior([w,gamma]))


# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================



""" Module with the functionality for evaluating the InferPy models """



import inferpy as inf
import edward as ed
from six import iteritems




ALLOWED_METRICS = ['binary_accuracy',
'categorical_accuracy',
'sparse_categorical_accuracy',
'log_loss',
'binary_crossentropy',
'categorical_crossentropy',
'sparse_categorical_crossentropy',
'hinge',
'squared_hinge',
'mse',
'MSE',
'mean_squared_error',
'mae',
'MAE',
'mean_absolute_error',
'mape',
'MAPE',
'mean_absolute_percentage_error',
'msle',
'MSLE',
'mean_squared_logarithmic_error',
'poisson',
'cosine',
'cosine_proximity',
'log_lik',
'log_likelihood']

""" List with all the allowed metrics for evaluation """



def evaluate(metrics, data, n_samples=500, output_key=None, seed=None):

    """ Evaluate a fitted inferpy model using a set of metrics. This function
    encapsulates the equivalent Edward one.

    Args:
        metrics: list of str indicating the metrics or sccore functions to be used.


    An example of use:

    .. literalinclude:: ../../examples/evaluate.py
       :language: python
       :lines: 52,53


    Returns:
        list of float or float: A list of evaluations or a single evaluation.


    Raises:
        NotImplementedError: If an input metric does not match an implemented metric.

    """

    data_ed = {}

    for (key, value) in iteritems(data):
        data_ed.update(
            {key.dist if isinstance(key, inf.models.RandomVariable) else key :
                 value.dist if isinstance(value, inf.models.RandomVariable) else value})

    output_key_ed = output_key.dist if isinstance(output_key, inf.models.RandomVariable) else output_key

    return ed.evaluate(metrics, data_ed, n_samples, output_key_ed, seed)


# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Module with useful definitions to be used in runtime
"""


import tensorflow as tf
from inferpy.util.wrappers import singleton
import edward as ed




@singleton
class Runtime():
    #def __init__(self):
    tf_run_default = True

    compact_param_str = True


    tf_sess = ed.get_session()
    init_g = tf.global_variables_initializer()
    init_l = tf.local_variables_initializer()
    tf_sess.run(init_g)
    tf_sess.run(init_l)


def get_session():
    """ Get the default tensorflow session """

    return Runtime.tf_sess


















import edward as ed
import inferpy as inf
from inferpy.models import Normal, Bernoulli, Categorical

d, N =  10, 500

# model definition
with inf.ProbModel() as m:

    #define the weights
    w0 = Normal(0,1)
    w = Normal(0, 1, dim=d)

    # define the generative model
    with inf.replicate(size=N):
        x = Normal(0, 1, observed=True, dim=d)
        y = Bernoulli(logits=w0+inf.dot(x, w), observed=True)


# toy data generation
x_train = Normal(loc=0, scale=1, dim=d).sample(N)
y_train = Bernoulli(probs=0.4).sample(N)
data = {x.name: x_train, y.name: y_train}

# compile and fit the model with training data
m.compile()
m.fit(data)

print(m.posterior([w, w0]))







import edward as ed
import inferpy as inf


#### learning a 1-dim parameter from 1-dim data


N = 50
sampling_mean = [30.]
sess = ed.util.get_session()


with inf.ProbModel() as m:

    theta = inf.models.Normal(loc=0., scale=1.)

    with inf.replicate(size=N):
        x = inf.models.Normal(loc=theta, scale=1., observed=True)

m.compile()

x_train = inf.models.Normal(loc=sampling_mean, scale=1.).sample(N)
data = {x.name : x_train}


m.fit(data)

m.posterior(theta).loc[0]  #29.017122


#### learning a 2 parameters of 1-dim from 2-dim data

N = 50
sampling_mean = [30., 10.]
sess = ed.util.get_session()


with inf.ProbModel() as m:

    theta1 = inf.models.Normal(loc=0., scale=1., dim=1)
    theta2 = inf.models.Normal(loc=0., scale=1., dim=1)


    with inf.replicate(size=N):
        x = inf.models.Normal(loc=[theta1, theta2], scale=1., observed=True)




m.compile()


x_train = inf.models.Normal(loc=sampling_mean, scale=1.).sample(N)
data = {x.name : x_train}


m.fit(data)

m.posterior(theta1).loc
m.posterior(theta2).loc




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

""" This module implements all the methods for definition of each type of random variable.
 For further details about all the supported distributions,
 see `Guide to Building Probabilistic Models <../notes/guidemodels.html#supported-probability-distributions>`_ .

"""


from inferpy.util import tf_run_wrapper

from inferpy.models.params import *

import six

import inspect

from inferpy.models import RandomVariable





######

CLASS_NAME = "class_name"
PARAMS = "params"
BASE_CLASS_NAME = "base_class_name"
NDIM = "ndim"
IS_SIMPLE="is_simple"
PROPS="properties"


def __add_property(cls, name, call_name = None, from_dist=False, private=True):

    call_name = name if call_name == None else call_name
    call_name = "__"+call_name if private and not from_dist else call_name

    if from_dist:
        @tf_run_wrapper
        def param(self):
            """Distribution parameter for the mean."""

            out = self.dist
            for m in call_name.split("."):
                out = getattr(out, m)

            return out


    else:
        def param(self):
            """Distribution parameter for the mean."""

            return getattr(self, call_name)


    cls.__perinstance = True
    param.__doc__ = "property for " + name
    param.__name__ = name

    setattr(cls, param.__name__, property(param))



def __add_constructor(cls, class_name, base_class_name, params, is_simple):

    def constructor(self,*args, **kwargs):

        param_dist = {}
        args_list = list(args)
        dist = None
        observed = kwargs.get("observed") if kwargs.get("observed") != None else False


        if len(args)+len(kwargs)>0:


            param_dim = kwargs.get("dim")
            param_batch = kwargs.get("batches")

            rep = None

            if param_batch != None:
                rep = inf.replicate(size=np.prod(param_batch))
                rep.__enter__()


            param_list = ParamList(params,args_list,kwargs,is_simple,param_dim=param_dim)

            if not param_list.is_empty():

                param_list.check_params()
                param_dist = param_list.get_reshaped_param_dict()

                ## Build the underliying tf object

                validate_args = kwargs.get("validate_args") if  kwargs.get("validate_args") != None else False
                allow_nan_stats = kwargs.get("allow_nan_stats") if  kwargs.get("allow_nan_stats") != None else True
                name = kwargs.get("name") if kwargs.get("name") != None else "inf/"+class_name


                dist = getattr(ed.models, class_name)(name=name, validate_args= validate_args, allow_nan_stats=allow_nan_stats, **param_dist)


            if rep != None:
                rep.exit()


        super(self.__class__, self).__init__(dist, observed=observed)




    constructor.__doc__ = "constructor for "+class_name
    constructor.__name__ = "__init__"
    setattr(cls, constructor.__name__, constructor)

def __add_repr(cls, class_name, params):

    def repr(self):

        if self.base_object != None:
            s = ", ".join([p+"="+inf.util.np_str(getattr(self,p)) for p in params])


            return "<inferpy.models."+class_name+" "+self.name+", "+s+", shape="+str(self.shape)+" >"
        else:
            return ""


    repr.__doc__ = "__repr__ for "+class_name
    repr.__name__ = "__repr__"
    setattr(cls, repr.__name__, repr)


### general method for defining a random variable class ###
def def_random_variable(var):

    # only the class_name is indicated
    if isinstance(var, six.string_types):
        v = {CLASS_NAME: var}
    else:
        v = var


    # check the name of the encapsulated Edward class
    if not BASE_CLASS_NAME in v:
        v.update({BASE_CLASS_NAME : v.get(CLASS_NAME)})

    # if the parameters (e.g., loc, scale, rate, etc.) are not indicated
    # these are obtained from tf.contrib.distributions
    if not PARAMS in v:

        lst = tf.distributions._allowed_symbols

        if v.get(BASE_CLASS_NAME) in lst:
            init_f = getattr(getattr(tf.contrib.distributions, v.get(BASE_CLASS_NAME)), "__init__")
        else:
            init_f = getattr(getattr(ed.models, v.get(BASE_CLASS_NAME)), "__init__")

        sig = inspect.getargspec(init_f)
        v.update({PARAMS: [x for x in sig.args ]})


    # remove non-specific parameters of the random variable
    v.update({PARAMS: [x for x in v.get(PARAMS) if x not in ['self', 'validate_args', 'allow_nan_stats', 'name', 'dtype']]})


    if not IS_SIMPLE in v:
        v.update({IS_SIMPLE : {}})


    # starts the definition of the new class
    newclass = type(v.get(CLASS_NAME), (RandomVariable,),{})


    # Add the properties
    if not PROPS in v:
        props = {}
        for p in v.get(PARAMS):
            if p not in ['self', 'validate_args', 'allow_nan_stats', 'name', 'dtype']:
                props.update({p:p})

        v.update({PROPS:props})
    for p, prop_call in six.iteritems(v.get(PROPS)):
        __add_property(newclass,p, call_name=prop_call, from_dist=True)


    # Add the constructor
    __add_constructor(newclass, v.get(CLASS_NAME), v.get(BASE_CLASS_NAME), v.get(PARAMS), v.get(IS_SIMPLE))
    __add_repr(newclass, v.get(CLASS_NAME), v.get(PARAMS))

    # set a static list with the name of its parameters
    newclass.PARAMS = v.get(PARAMS)


    globals()[newclass.__name__] = newclass



####

class Normal(RandomVariable):
    def __init__(self, loc=0, scale=1,
                 validate_args=False,
                 allow_nan_stats=True,
                 dim=None, batches=None, observed=False, name="Normal"):
        self.loc = loc
        self.scale = scale


class Beta(RandomVariable):
    def __init__(
            self,
            concentration1=None,
            concentration0=None,
            validate_args=False,
            allow_nan_stats=True,
            observed=False,
            dim=None, batches=None,
            name='Beta'):
        self.concentration1 = concentration1
        self.concentration0 = concentration0



class Exponential(RandomVariable):
    def __init__(
            self,
            rate,
            validate_args=False,
            allow_nan_stats=True,
            observed = False,
            dim = None,
            name='Exponential'):
        self.rate = rate

class Uniform(RandomVariable):
    def __init__(
            self,
            low=None,
            high=None,
            validate_args=False,
            allow_nan_stats=True,
            name='Uniform',
            observed=False,
            dim=None, batches=None):
        self.low = low

class Poisson(RandomVariable):
    def __init__(
            self,
            rate,
            validate_args=False,
            allow_nan_stats=True,
            name='Poisson',
            observed=False,
            dim=None, batches=None):
        self.rate = rate

class Categorical(RandomVariable):
    def __init__(
            self,
            logits=None,
            probs=None,
            validate_args=False,
            allow_nan_stats=True,
            name='Categorical',
            observed=False,
            dim=None, batches=None):
        self.default_logits = logits
        self.default_probs = probs

class Multinomial(RandomVariable):
    def __init__(
            self,
            total_count=None,
            logits=None,
            probs=None,
            validate_args=False,
            allow_nan_stats=True,
            name='Categorical',
            observed=False,
            dim=None, batches=None):
        self.logits = logits
        self.probs = probs
        self.total_count = None

class Dirichlet(RandomVariable):
    def __init__(self,
            concentration,
            validate_args=False,
            allow_nan_stats=True,
            name='Dirichlet',
            observed=False,
            dim=None, batches=None):
        self.concentration=concentration

class Gamma(RandomVariable):
    def __init__(
            self,
            alpha, beta,
            validate_args=False,
            allow_nan_stats=True,
            observed=False,
            dim=None, batches=None,
            name='Gamma'):
        self.alpha = alpha
        self.beta = beta

class InverseGamma(RandomVariable):
    def __init__(
            self,
            alpha, beta,
            validate_args=False,
            allow_nan_stats=True,
            observed=False,
            dim=None, batches=None,
            name='InverseGamma'):
        self.alpha = alpha
        self.beta = beta

class Bernoulli(RandomVariable):
    def __init__(
            self,
            logits=None,
            probs=None,
            validate_args=False,
            allow_nan_stats=True,
            name='Bernoulli',
            observed=False,
            dim=None, batches=None):
        self.default_logits = logits
        self.default_probs = probs

class Laplace(RandomVariable):
    def __init__(self, loc, scale,
                 validate_args=False,
                 allow_nan_stats=True,
                 dim=None, batches=None, observed=False, name="Laplace"):
        self.loc = loc
        self.scale = scale

class MultivariateNormalDiag(RandomVariable):
    def __init__(self,
                 loc=None,
                 scale_diag=None,
                 validate_args=False,
                 allow_nan_stats=True,
                 name="MultivariateNormalDiag"):
        self.loc = loc
        self.scale_diag = scale_diag


####### run-time definition of random variables #########

SIMPLE_VARS = ["Normal","Beta", "Exponential","Uniform","Poisson", "Gamma", "Laplace",
               {CLASS_NAME : "InverseGamma", PARAMS : ['concentration', 'rate', 'self', 'validate_args', 'allow_nan_stats', 'name', 'dtype']}]


for v in SIMPLE_VARS:
    def_random_variable(v)
    g = globals()



NON_SIMPLE_VARS = [{CLASS_NAME : "Categorical", IS_SIMPLE : {"probs" : False, "logits": False}},
                   {CLASS_NAME: "Multinomial", IS_SIMPLE: {"total_count":True,"probs": False, "logits": False}},
                   {CLASS_NAME: "Dirichlet", IS_SIMPLE: {"concentration": False}},
                   {CLASS_NAME : "Bernoulli", IS_SIMPLE : {"probs" : True, "logits": True}},
                   {CLASS_NAME: "MultivariateNormalDiag",
                    IS_SIMPLE: {"loc": False, "scale_diag": False},
                    PARAMS: ['loc', 'scale_diag'],
                    PROPS:{'loc': 'loc', 'scale_diag': 'scale.diag'}
                    }

                   ]

for v in NON_SIMPLE_VARS:
    def_random_variable(v)


#####




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


""" Module implementing some useful operations over tensors and random variables """




import numpy as np
import inferpy.models
import tensorflow as tf
import collections
import six




def param_to_tf(x):

    """ Transforms either a scalar or a random variable into a Tensor"""


    if np.isscalar(x):
        return tf.constant(x, dtype="float32")
    elif isinstance(x, inferpy.models.RandomVariable):
        return x.base_object
    else:
        raise ValueError("wrong input value in param_to_tf")



def case_states(var, d, default=None, exclusive=True, strict=False, name='case'):

    """ Control flow operation depending of the outcome of a discrete variable.

    Internally, the operation tensorflow.case is invoked. Unlike the tensorflow operation, this one
    accepts InferPy variables as input parameters.

    Args:
        var: Control InferPy discrete random variable.
        d : dictionary where the keys are each of the possible values of control variable
        and the values are returning tensors for each case.
        exclusive: True iff at most one case is allowed to evaluate to True.
        name: name of the resulting tensor.

    Return:
        Tensor implementing the case operation. This is the output of the operation
        tensorflow.case internally invoked.


    """

    out_d = {}

    if not isinstance(var, inferpy.models.RandomVariable):
        var = inferpy.models.Deterministic(var)


    def f(p): return tf.constant(p)


    for s, p in six.iteritems(d):

        out_d.update({tf.reduce_all(tf.equal(var.dist, tf.constant(s))): (lambda pp : lambda: f(pp))(p)})

    return tf.case(out_d, default=default, exclusive=exclusive,strict=strict,name=name)



def case(d, default=None, exclusive=True, strict=False, name='case'):

    """ Control flow operation depending of the outcome of a tensor. Any expression
    in tensorflow giving as a result a boolean is allowed as condition.

    Internally, the operation tensorflow.case is invoked. Unlike the tensorflow operation, this one
    accepts InferPy variables as input parameters.

    Args:
        d : dictionary where the keys are the conditions (i.e. boolean tensor).
        exclusive: True iff at most one case is allowed to evaluate to True.
        name: name of the resulting tensor.

    Return:
        Tensor implementing the case operation. This is the output of the operation
        tensorflow.case internally invoked.


    """


    out_d = {}

    def f(p): return tf.constant(p)

    for c, p in six.iteritems(d):

        out_d.update({tf.reduce_all(tf.equal(c.base_object, True)): (lambda pp : lambda: f(pp))(p)})


    if default != None:
        default = (lambda pp : lambda: f(pp))(default)

    return tf.case(out_d, default=default, exclusive=exclusive,strict=strict,name=name)



def gather(
        params,
        indices,
        validate_indices=None,
        name=None,
        axis=0 ):
    """ Operation for selecting some of the items in a tensor.

    Internally, the operation tensorflow.gather is invoked. Unlike the tensorflow operation, this one
    accepts InferPy variables as input parameters.

    Args:
        params: A Tensor. The tensor from which to gather values. Must be at least rank axis + 1.
        indices: A Tensor. Must be one of the following types: int32, int64. Index tensor. Must be in range
        [0, params.shape[axis]).
        axis: A Tensor. Must be one of the following types: int32, int64. The axis in params to gather indices
        from. Defaults to the first dimension. Supports negative indexes.
        name: A name for the operation (optional).



    Return:
        A Tensor. Has the same type as params.. This is the output of the operation
        tensorflow.gather internally invoked.


    """


    tf_params = params.base_object if isinstance(params, inferpy.models.RandomVariable)==True else params
    tf_indices = indices.base_object if isinstance(indices, inferpy.models.RandomVariable) == True else indices

    return  tf.gather(tf_params, tf_indices, validate_indices, name, axis)


def matmul(
        a,
        b,
        transpose_a=False,
        transpose_b=False,
        adjoint_a=False,
        adjoint_b=False,
        a_is_sparse=False,
        b_is_sparse=False,
        name=None):


    """ Matrix multiplication.

    Input objects may be tensors but also InferPy variables.

    Args:
        a: Tensor of type float16, float32, float64, int32, complex64, complex128 and rank > 1.
        b: Tensor with same type and rank as a.
        transpose_a: If True, a is transposed before multiplication.
        transpose_b: If True, b is transposed before multiplication.
        adjoint_a: If True, a is conjugated and transposed before multiplication.
        adjoint_b: If True, b is conjugated and transposed before multiplication.
        a_is_sparse: If True, a is treated as a sparse matrix.
        b_is_sparse: If True, b is treated as a sparse matrix.
        name: Name for the operation (optional).

    Retruns:
        An InferPy variable of type Deterministic encapsulating the resulting tensor
        of the multiplications.


    """

    res = inferpy.models.Deterministic()


    a_shape = shape_to_list(a)
    b_shape = shape_to_list(b)


    if isinstance(a, inferpy.models.RandomVariable):
        a = a.base_object

    if isinstance(b, inferpy.models.RandomVariable):
        b = b.base_object


    a = a if len(a_shape) > 1 else tf.reshape(a, [1] + a_shape)
    b = b if len(b_shape) > 1 else tf.reshape(b, [1] + b_shape)

    res.base_object = tf.matmul(a, b, transpose_a, transpose_b, adjoint_a, adjoint_b, a_is_sparse, b_is_sparse, name)

    return res


def dot(x,y):

    """ Compute dot product between an InferPy or Tensor object. The number of batches N equal to 1
    for one of them, and higher for the other one.

     If necessarily, the order of the operands may be changed.

     Args:
         x: first operand. This could be an InferPy variable, a Tensor, a numpy object or a numeric Python list.
         x: second operand. This could be an InferPy variable, a Tensor, a numpy object or a numeric Python list.


    Retruns:
        An InferPy variable of type Deterministic encapsulating the resulting tensor
        of the multiplications.

     """


    x_shape = shape_to_list(x)
    y_shape = shape_to_list(y)

    if len(x_shape) == 1 and len(y_shape)==2:

        a = y
        b = x

    elif len(x_shape) == 2 and len(y_shape) == 1:
        a = x
        b = y


    else:
        raise ValueError("Wrong dimensions")


    return matmul(a, b, transpose_b=True)



def shape_to_list(a):

    """ Transforms the shape of an object into a list

    Args:
        a : object whose shape will be transformed. This could be an InferPy variable, a Tensor, a numpy object or a numeric Python list.

    """

    if isinstance(a, inferpy.models.RandomVariable):
        a_shape = a.shape
    elif isinstance(a, np.ndarray):
        a_shape = list(a.shape)
    elif isinstance(a, list):
        a_shape = list(np.shape(a))
    elif isinstance(a, tf.Tensor):
        a_shape = a._shape_as_list()
    else:
        raise ValueError("Wrong input type "+a)

    return a_shape


def fix_shape(s):

    """ Transforms a shape list into a standard InferPy shape format. """

    ret = []

    for i in range(0,len(s)):
        if i in [0, len(s)-1] or s[i] != 1:
            ret.append(s[i])

    if len(ret) == 0:
        return [1]

    return ret









import edward as ed

import tensorflow as tf

import tensorflow.contrib.distributions as ds

import inferpy as inf


[c.__name__ for c in tf.contrib.distributions.Distribution.__subclasses__()]

getattr(tf.contrib.distributions, "PointMass")
getattr(getattr(ed.models, "PointMass"), "__init__")(params=1.)


d = inf.models.PointMass(params=1.)

d.base_object


import edward as ed
import inferpy as inf
from inferpy.models import Normal
import numpy as np

import tensorflow as tf

d, N =  5, 1000

with inf.replicate(size=N):
    x = inf.models.MultivariateNormalDiag(
        loc=[1., -1],
        scale_diag=[1, 2.],dim=5
    )

with inf.replicate(size=N):
    x = inf.models.Normal(0,1,dim=d)



x = inf.models.MultivariateNormalDiag(
    loc=[1., -1],
    scale_diag=[1, 2.],dim=[5,10]
)



x = inf.models.MultivariateNormalDiag(
    loc=[1., -1],
    scale_diag=[1, 2.], dim=[100,100], batch = 5
)



x.event_shape
x.dim
x.batches

x.shape

x.loc.shape


with tf.Session() as sess:
    print(x.dist.shape)







import inferpy as inf


with inf.replicate(size=50,name="A"):
    # Define some random variables here
    print("Variable replicated " + str(inf.replicate.get_total_size()) + " times")


with inf.replicate(size=10):
    # Define some random variables here
    print("Variable replicated " + str(inf.replicate.get_total_size()) + " times")

    with inf.replicate(size=2, name="C"):
        # Define some random variables here
        print("Variable replicated " + str(inf.replicate.get_total_size()) + " times")
        print(inf.replicate.in_replicate())

# Define some random variables here
print("Variable replicated " + str(inf.replicate.get_total_size()) + " times")


### existing replicate construct can be reused.
### This is done by ommiting the size argument and only setting the name with the name of
### an existing one.


with inf.replicate(name="A"):
    with inf.replicate(name="C"):
        # Define some random variables here
        print("Variable replicated " + str(inf.replicate.get_total_size()) + " times")



inf.replicate.get_active_replicate()


import inferpy as inf
from inferpy.models import Normal, Beta, Categorical, Deterministic



d, N =  10, 200


#define the weights
w0 = Normal(0,1)
with inf.replicate(size=d):
    w = Normal(0, 1)


p = Beta(1,1)

# define the generative model
with inf.replicate(size=N):
    x = Normal(0, 1, observed=True, dim=d)
    y0 = Normal(inf.matmul(x,w), 1.0, observed=True)

    h = Categorical(probs=[p, 1-p])
    y1 = Deterministic(1.)

    # not working until issue #58 is solved
    y = Deterministic(inf.case({h.equal(0) : y0, h.equal(1) : y1}), observed = True)


h = Categorical(probs=[0.2,0.8])
h.probs
h.sample()





# toy data generation
x_train = Normal(loc=0, scale=1, dim=d).sample(N)
y_train = Normal(loc=5, scale=1, dim=1).sample(N)
data = {x.name: x_train, y.name: y_train}


# compile and fit the model with training data
m.compile()
m.fit(data)

m.posterior([w, w0])


x.dist.__matmul__(w.dist)

import tensorflow as tf

tf.matmul(x.dist,w.dist, transpose_a=False, transpose_b=True)


import inferpy as inf
from inferpy.models import Normal



with inf.ProbModel() as m:

    x = Normal(loc=1., scale=1., name="x", observed=True)
    y = Normal(loc=x, scale=1., dim=3, name="y")


# print the list of variables
print(m.varlist)
print(m.latent_vars)
print(m.observed_vars)

# get a sample

m_sample = m.sample()



# compute the log_prob for each element in the sample
print(m.log_prob(m_sample))

# compute the sum of the log_prob
print(m.sum_log_prob(m_sample))




### alternative definition

x2 = Normal(loc=1., scale=1., name="x2", observed=True)
y2 = Normal(loc=x, scale=1., dim=3, name="y2")

m2 = inf.ProbModel(varlist=[x2,y2])




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Module with the replication functionality.
"""

from functools import reduce

class replicate():
    """Class implementing the Plateau notation

        The plateau notation is used to replicate the random variables contained
        within this construct. Every replicated variable is conditionally idependent
        given the previous random variables (if any) defined outside the with statement.
        The ``with inf.replicate(size = N)`` sintaxis is used to replicate N times the
        contained definitions. For example:

        .. literalinclude:: ../../examples/replicate_nested.py


        The number of times that indicated with input argument ``size``.
        Note that nested replicate constructs can be defined as well. At any moment,
        the product of all the nested replicate constructs can be obtained by
        invoking the static method ``get_total_size()``.

        .. note::

            Defining a variable inside the construct replicate with size equal to 1, that is,
            ``inf.replicate(size=1)`` is equivalent to defining outside any replicate
            construct.


        """


    __active_rep = []
    __all_rep = []
    __refrep = None



    def __init__(self,size=None,name="rep"):
        """Initializes the replicate construct

        Args:
            size (int): number of times that the variables contained are replicated.

        """

        ### creating a new replicate
        if(size != None):
            self.size = size
            self.varlist=[]

            if(replicate.exists(name)):
                if(name != "rep"):
                    raise ValueError("ERROR: existing replicate construct named "+name)
                else:
                    i = len(replicate.get_all_replicate())+1
                    if(i>1):
                        name = name+str(i)


            else:
                if(name.startswith("rep_")):
                    raise ValueError("ERROR: invalid replicate name "+name)


            self.__name = name
        else:
            ## check that this is not active
            if(replicate.is_active(name)):
                raise ValueError("A replicate construct cannot be compounded with itself")

            # when this value is set, in the enter method, current object will be discarded
            # and the existing construct will be activated
            self.__refrep=name





    @staticmethod
    def __get_all_sizes():
        return [1]+[r.size for r in replicate.__active_rep]


    def __enter__(self):


        # refering to an existing replicate
        if (self.__refrep != None):
            r = replicate.get(self.__refrep)
            replicate.__active_rep.append(r)
            return r

        # normal behaivour

        replicate.__active_rep.append(self)
        replicate.__all_rep.append(self)
        return self



    def __exit__(self, exc_type, exc_val, exc_tb):

        self.exit()

    def exit(self):
        replicate.__active_rep.pop()



    @staticmethod
    def get_total_size():
        """Static method that returns the product of the sizes of all the nested replicate constructs

        Returns:
            Integer with the product of sizes

        """
        if len(replicate.__get_all_sizes()) == 1:
            return 1
        return reduce(lambda x, y: x * y, replicate.__get_all_sizes())

    @staticmethod
    def print_total_size():
        """Static that prints the total size
        """
        print("total size: " + str(replicate.get_total_size()))

    @staticmethod
    def in_replicate():
        """Check if a replicate construct has been initialized

        Returns:
             True if the method is inside a construct replicate (of size different to 1).
             Otherwise False is return
        """
        return len(replicate.__active_rep)>0



    # static methods
    @staticmethod
    def get_active_replicate():

        """ Return the active replicate defined with the construct 'with' """

        if replicate.in_replicate():
            return replicate.__active_rep
        return []

    @staticmethod
    def get_all_replicate():
        return replicate.__all_rep

    @staticmethod
    def exists(name):
        return name in [r.name for r in replicate.get_all_replicate()]

    @staticmethod
    def is_active(name):
        return name in [r.name for r in replicate.get_active_replicate()]

    @staticmethod
    def get(name):
        return [r for r in replicate.get_all_replicate() if r.name==name][0]


    @staticmethod
    def delete_all():
        replicate.__active_rep = []
        replicate.__all_rep = []



    ##### properties and setters ########

    @property
    def name(self):
        return self.__name




# -*- coding: utf-8 -*-
#


__version__ = '0.2.1rc1'
VERSION = __version__


from inferpy import criticism
from inferpy import models
from inferpy import util
from inferpy import inferences

#from inferpy.inferences import inference


# Direct imports for convenience
from inferpy.criticism.evaluate import *
from inferpy.util.runtime import get_session
from inferpy.models.prob_model import *
from inferpy.models.replicate import *

from inferpy.inferences.inference import INF_METHODS
from inferpy.inferences.inference import INF_METHODS_ALIAS
from inferpy.inferences.qmodel import *



## basic funcitons

from inferpy.util.ops import *








import inferpy as inf
import edward as ed
import tensorflow as tf
import numpy as np


sess = ed.get_session()



k=2

N = 1000 # number of observations
# model definition
with inf.ProbModel() as m:

    # observed variable
    with inf.replicate(size=N):
        x = inf.models.Normal(loc=10, scale=1, observed=True, name="x")
        h = inf.models.Normal(loc=k*x, scale=1, observed=False, name="h")



x_train = inf.models.Normal(loc=30/10, scale=10, dim=1).sample(N)


#m.compile()
#m.fit({x.name: x_train})

qh = inf.Qmodel.new_qvar(h, check_observed=False)


x_test = x_train
h_test = k * x_train


#not working
ed.KLqp({h.dist : qh.dist}, data={x.dist : x_test}).run(n_iter=10000)


# this works
map = ed.MAP([h.dist], data={x.dist : x_test})
map.run()
h_pred = sess.run(map.latent_vars.values()[0])


# this works
ed.KLqp({ed.copy(h.dist, {x.dist : x_test}) : qh.dist}, data={}).run()


# not working
ed.KLpq({h.dist : qh.dist}, data={x.dist : x_test}).run()


# works
ed.ReparameterizationKLqp({h.dist : qh.dist}, data={x.dist : x_test}).run()


# not working
ed.ReparameterizationKLKLqp({h.dist : qh.dist}, data={x.dist : x_test}).run()


# works
ed.ReparameterizationEntropyKLqp({h.dist : qh.dist}, data={x.dist : x_test}).run()

#not working
ed.ScoreKLqp({h.dist : qh.dist}, data={x.dist : x_test}).run()

# not working
ed.ScoreKLKLqp({h.dist : qh.dist}, data={x.dist : x_test}).run()


# not working
ed.ScoreEntropyKLqp({h.dist : qh.dist}, data={x.dist : x_test}).run()




h_pred = qh.loc
x_train

err = np.sum(np.power(h_pred - h_test, 2)) / N


for i in range(0,5):
    print(str(h_test[-i])+" -- "+str(h_pred[-i]))



#np.mean(m.posterior(h).loc)


import inferpy as inf
import tensorflow as tf
import edward as ed
import numpy as np
import tensorflow as tf


## to do: check estructure and minimal
## try an algo
## informational pred of D

class Chance(inf.models.Categorical):

    def __init__(self, probs, parents = None, name="Chance"):
        self.__probs = probs
        self.__parents = parents

        parents = [] if parents is None else parents

        if parents not in ([], np.array([])):
            p = inf.case_states(parents, probs)
        else:
            p = probs


        self.dist = inf.models.Categorical(probs=p, name=name).dist
        self.observed = False

    @property
    def probs(self):
        return self.__probs

    @property
    def parents(self):
        return self.__parents



class Decision(inf.models.Categorical):

    def __init__(self, num_options=2, name="Decision"):
        self.__num_options = num_options

        self.dist = inf.models.Categorical(logits=np.zeros(num_options), name=name).dist
        self.observed = False

    @property
    def num_options(self):
        return self.__num_options





class Utility(inf.models.Categorical):

    def __init__(self, utils=None, parents = None, name="Util"):

        if utils != None:

            self.__utils = utils
            self.__parents = parents



            parents = [] if parents is None else parents

            # Cooper transformation

            k2 = - np.min(utils.values())
            k1 = np.max(utils.values()) + k2

            self.k2 = k2
            self.k1 = k1

            import six

            pv = {}
            for (k, v) in six.iteritems(utils):
                p = (1.0 * np.abs(v + k2) / k1)[0]
                pv.update({k: [1-p, p]})

            p = inf.case_states(parents, pv)

            self.dist = inf.models.Categorical(probs=p, name=name).dist
            self.observed = False


    @property
    def utils(self):
        return self.__utils

    @property
    def parents(self):
        return self.__parents





x = Chance([0.5,0.5])

d = Decision(num_options=2)

#y = Chance(probs = { (0, 0): [1.0, 0.0, 0.0, 0.0],
#                     (0, 1): [0.0, 1.0, 0.0, 0.0],
#                     (1, 0): [0.0, 0.0, 1.0, 0.0],
#                     (1, 1): [0.0, 0.0, 0.0, 1.0]}, parents=[x,d])

sess = inf.get_session()

v = Utility(utils={(0, 0): [10.0],
                  (0, 1): [0.0],
                  (1, 0): [50.0],
                  (1, 1): [-5.0]}, parents=[x, d])




# P(v=1)

def id():
    return str(inf.models.Uniform(0,10000000).sample(1).astype(int)[0,0])


v.probs.shape

s = inf.get_session()

d_obs = inf.models.Categorical(probs=[1,0])
qv = inf.Qmodel.new_qvar(v, qvar_inf_module=inf.models, qvar_type="Categorical", check_observed=False)
#ed.inferences.ReparameterizationKLqp({v.dist : qv.dist}, {d.dist:d_obs.dist}).run(n_iter=1000)
ed.inferences.ReparameterizationKLqp({v.dist : qv.dist}, {d.dist:[0]}).run(n_iter=1000)

putil = (qv.probs/qv.probs.sum())
ud0 = v.k1 * putil[0,1] - v.k2


#ed.inferences.ReparameterizationKLqp({ed.copy(v.dist,{d.dist:d_obs.dist}) : qv.dist}, {}).run()

#ed(ed.copy(v.dist,{d.dist:d_obs.dist}).sample())





d_obs = inf.models.Categorical(probs=[0,1])
qv = inf.Qmodel.new_qvar(v, qvar_inf_module=inf.models, qvar_type="Categorical", check_observed=False)
ed.inferences.ReparameterizationKLqp({v.dist : qv.dist}, {d.dist:[1]}).run(n_iter=10000)
putil = (qv.probs/qv.probs.sum())

ud1 = v.k1 * putil[0,1] - v.k2

print([ud0,ud1])




####



a = Chance([0.5,0.5])

u = Utility(utils={0:[10], 1:[5]}, parents=[a])

qu = inf.Qmodel.new_qvar(u, qvar_inf_module=inf.models, qvar_type="Categorical", check_observed=False)
ed.inferences.ReparameterizationKLqp({u.dist : qu.dist}, {d.dist:[1]}).run(n_iter=1000)

qu.probs

pu = (qu.probs/qu.probs.sum())[0,1]

u_post = u.k1 * pu[0,1] - u.k2

import math
pu = u.k1 * (1/ (1 + math.exp(-qu.logits[0,1]))) - u.k2


with inf.replicate(size=10):
    c = inf.models.Categorical(logits=[0,0,0], dim=5)

c.event_shape


print(u_post)

qu






import inferpy as inf


inf.models.ALLOWED_VARS

# Bernoulli: binary distribution with probability p and 1-p

x = inf.models.Bernoulli(probs=0.5)

# or

x = inf.models.Bernoulli(logits=0)


#################   15
# Beta: continuous distribution defined in the interval [0,1]
# parametrized by two positive shape parameters, denoted alpha and beta

x = inf.models.Beta(concentration0=0.5, concentration1=0.5)

# or simply:

x = inf.models.Beta(0.5,0.5)


# Categorical: discrete distribution

x = inf.models.Categorical(probs=[0.5,0.5])

# or

x = inf.models.Categorical(logits=[0,0])


# Deterministic

x = inf.models.Deterministic(loc=5)

# or simply:

x = inf.models.Deterministic(5)

############# 43

# Dirichlet: continuous multivariate probability distributions parameterized by a vector
#  of positive reals. It is a multivariate generalization of the beta distribution. Dirichlet
# distributions are commonly used as prior distributions in Bayesian statistics

x = inf.models.Dirichlet(concentration=[5,1])

# or simply:

x = inf.models.Dirichlet([5,1])

#############55
# Exponential:
x = inf.models.Exponential(rate=1)

# or simply

x = inf.models.Exponential(1)

# Gamma:



x = inf.models.Gamma(concentration=3, rate=2)

# Inverse Gamma: NOT WORKING



x = inf.models.InverseGamma(concentration=3, rate=2)

###########75

# Laplace: the Laplace distribution is a continuous probability distribution


x = inf.models.Laplace(loc=0, scale=1)

# or simply

x = inf.models.Laplace(0,1)




# Multinomial

# The multinomial is a discrete distribution

x = inf.models.Multinomial(total_count=4, probs=[0.5,0.5])

# or

x = inf.models.Multinomial(total_count=4, logits=[0,0])



# Normal



x = inf.models.Normal(loc=0, scale=1)

# or

x = inf.models.Normal(0,1)


########112
# Poisson


x = inf.models.Poisson(rate=4)

# or

x = inf.models.Poisson(4)


####### 123

# Uniform


x = inf.models.Uniform(low=1, high=3)

# or

inf.models.Uniform(1,3)



#136



x = inf.models.MultivariateNormalDiag(
    loc=[1., -1],
    scale_diag=[1, 2.]
)



import tensorflow as tf
import os
import inferpy as inf
import numpy as np
import edward as ed
N=10
d=5


with inf.replicate(size=N):
    y = inf.models.Normal(loc=0, scale=1, dim=d) #shape=[N,d]


c = inf.models.Categorical(logits = np.zeros(5), name="c")
yc = inf.models.Normal(loc=2+y[0,c], scale=1)  #shape = [1]



y.dist

c.sample()

c.sample(tf_run=False)

"""

>>> y.dist

<ed.RandomVariable 'inf/Normal_1/' shape=(10, 5) dtype=float32>

>>> c.sample()

array([[3]], dtype=int32)

>>> c.sample(tf_run=False)

<tf.Tensor 'z/sample_2/Reshape_1:0' shape=(1, 1) dtype=int32>

"""

K, d, N, T = 3, 4, 1000, 5000


# toy data generation
x_train = np.vstack([inf.models.Normal(loc=0, scale=1, dim=d).sample(300),
                     inf.models.Normal(loc=10, scale=1, dim=d).sample(700)])


## model definition ##
with inf.ProbModel() as model:

    # prior distributions
    with inf.replicate(size=K):
        mu = inf.models.Normal(loc=0, scale=1, dim=d)
        sigma = inf.models.InverseGamma(1, 1, dim=d)
    p = inf.models.Dirichlet(np.ones(K)/K)

    # define the generative model
    with inf.replicate(size=N):
        z = inf.models.Categorical(probs = p)
        x = inf.models.Normal(mu[z], sigma[z],
                              observed=True, dim=d)

# compile and fit the model with training data
data = {x: x_train}
model.compile(infMethod="MCMC")
model.fit(data)
# print the posterior
print(model.posterior(mu))

######## Edward ##########   71



## model definition ##
# prior distributions
p = ed.models.Dirichlet(concentration=tf.ones(K)/K)
mu = ed.models.Normal(0.0, 1.0, sample_shape=[K, d])
sigma = ed.models.InverseGamma(concentration=1.0, rate=1.0,
                               sample_shape=[K, d])
# define the generative model
z = ed.models.Categorical(logits=tf.log(p) - tf.log(1.0 - p),
                          sample_shape=N)
x = ed.models.Normal(loc=tf.gather(mu, z), scale=tf.gather(sigma, z))



# compile and fit the model with training data
qp = ed.models.Empirical(params = tf.get_variable( "qp/prm", [T,K],
                    initializer =tf.constant_initializer(1.0 / K)))
qmu = ed.models.Empirical( params= tf.get_variable("qmu/prm",[T,K,d],
                    initializer = tf.zeros_initializer()))
qsigma = ed.models.Empirical(params= tf.get_variable("qsigma/prm", [T,K,d],
                    initializer = tf.ones_initializer()))
qz = ed.models.Empirical( params= tf.get_variable("qz/prm", [T, N],
                    initializer= tf.zeros_initializer(),
                    dtype=tf.int32))

gp = ed.models.Dirichlet(concentration=tf.ones(K))
gmu = ed.models.Normal(loc=tf.ones([K,d]), scale=tf.ones([K,d]))
gsigma = ed.models.InverseGamma(concentration= tf.ones([K,d]), rate=tf.ones([K,d]))
gz = ed.models.Categorical(logits=tf.zeros([N, K]))

inference = ed.MetropolisHastings(
    latent_vars={p: qp, mu: qmu,
                 sigma: qsigma, z: qz},
    proposal_vars={p: gp, mu: gmu,
                   sigma: gsigma, z: gz},
    data={x: x_train})
inference.run()

# print the posterior
print(qmu.params.eval())



######### comp  117

qmu = ed.models.Empirical(params=tf.get_variable(
    "qmu/prm", [T,K,d],
    initializer=tf.zeros_initializer()))
gmu = ed.models.Normal(loc=tf.ones([K,d]),
                       scale=tf.ones([K,d]))



# -*- coding: utf-8 -*-
#


from setuptools import setup, find_packages

import os
import re


here = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(here, 'inferpy', '__init__.py')) as f:
    txt = f.read()

try:
    version = re.findall(r"^__version__ = '([^']+)'\r?$",
                         txt, re.M)[0]
except IndexError:
    raise RuntimeError('Unable to determine version.')



with open("project_description.md", "r") as fh:
    long_description = fh.read()


setup(
    name='inferpy',
    version=version,
    description='Probabilistic modeling with Tensorflow made easy',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Andrés R. Masegosa, Rafael Cabañas',
    author_email="andresma@ual.es, rcabanas@ual.es",
    install_requires=['tensorflow >= 1.5, <1.8', 'numpy>=1.14', 'edward==1.3.5', 'pandas>0.15.0'],
    extras_require={
        'tensorflow with gpu': ['tensorflow-gpu  >= 1.5, <1.8'],
        'visualization': ['matplotlib>=1.3',
                          'pillow>=3.4.2',
                          'seaborn>=0.3.1']},
    tests_require=['pytest', 'pytest-pep8'],
    url='http://inferpy.readthedocs.io',
    download_url='https://github.com/PGMLabSpain/InferPy/archive/' + version + '.tar.gz',
    keywords='machine learning statistics probabilistic programming tensorflow edward',
    license='Apache License 2.0',
    classifiers=['Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: Apache Software License',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: Microsoft :: Windows',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.4'],
    packages=find_packages(),
)









# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Module with the probabilistic model functionality.
"""

import edward as ed
import tensorflow as tf
from six import iteritems

import inferpy as inf


from inferpy.util import input_model_data
from inferpy.util import multishape
from inferpy.util import tf_run_wrapper


class ProbModel(object):
    """Class implementing a probabilistic model

    A probabilistic model defines a joint distribution over observed and latent variables. This
    class encapsulates all the functionality for making inference (and learning) in these models.

    An example of use:

    .. literalinclude:: ../../examples/prob_model_def.py


    This class can be used, for instance, for infering the parameters of some observed data:

    .. literalinclude:: ../../examples/simple_inference_params.py



    """



    __active_models = []
    """ list: static variable that contains the models defined by means
    of the construct 'with' that are active"""


    def __init__(self, varlist=None):
        """Initializes a probabilistic model

        Args:
            varlist: optional list with the variables in the model

        """


        if varlist==None:
            self.varlist = []
        else:
            self.varlist=varlist


        for d in self.varlist:
            if not ProbModel.compatible_var(d):
                raise ValueError("The input argument is not a list of RandomVariables")

        if ProbModel.is_active():
            raise inf.util.ScopeException("Nested probabilistic models cannot be defined")



        self.Q = None
        self.data = {}

        self.propagated = False

    # properties and setters

    @property
    def varlist(self):
        """ list of variables (observed and latent)"""
        return self.__varlist

    @varlist.setter
    def varlist(self,varlist):

        """ modifies the list of variables in the model"""

        self.__varlist = varlist
        self.reset_compilation()

    @property
    def observed_vars(self):

        """ list of observed variabels in the model"""

        vl = []
        for v in self.varlist:
            if v.observed:
                vl.append(v)
        return vl


    @property
    def latent_vars(self):

        """ list of latent (i.e., non-observed) variables in the model """

        vl = []
        for v in self.varlist:
            if v.observed==False:
                vl.append(v)
        return vl

    # other methods



    def compile(self, infMethod="KLqp", Q=None, proposal_vars=None):

        """ This method initializes the structures for making inference in the model."""


        # check if the infMethod is an alias
        if infMethod in inf.INF_METHODS_ALIAS.keys():
            infMethod = inf.INF_METHODS_ALIAS.get(infMethod)


        # check if the inference method exists
        if infMethod not in inf.INF_METHODS:
            raise ValueError("Unsupported inference method: "+infMethod)

        self.infMethod = infMethod

        self.Q = Q

        if self.Q == None:
            self.Q = inf.Qmodel.build_from_pmodel(self, self.infMethod=="MetropolisHastings")

        if proposal_vars==None and self.__inf_with_proposal_vars():
            self.proposal_vars = dict([(v.dist, v.dist.copy()) for v in self.latent_vars])


        self.propagated = False

    @input_model_data
    def fit(self, data):

        """ Assings data to the observed variables"""

        if self.is_compiled()==False:
            raise Exception("Error: the model is not compiled")

        self.data = {}

        for k, v in iteritems(data):
            new_k = self.get_var(k).dist if isinstance(k,str) else k.dist
            self.data.update({new_k : v})



        # prepare inference arguments
        inf_args = {"latent_vars":self.Q.dict}
        if self.__inf_with_proposal_vars():
            inf_args.update({"proposal_vars":self.proposal_vars})

        inf_args.update({"data": self.data})


        # inference
        self.inference = getattr(ed.inferences, self.infMethod)(**inf_args)


        self.inference.run()


        self.propagated = True

    @multishape
    def posterior(self, latent_var):

        """ Return the posterior distribution of some latent variables

            Args:
                latent_var: a single or a set of latent variables in the model

            Return:
                Random variable(s) of the same type than the prior distributions

        """

        if self.propagated == False:
            self.inference.run()
            self.propagated = True

        ed_post = self.inference.latent_vars.get(latent_var.dist)
        vartype = type(ed_post).__name__


        if vartype in inf.models.ALLOWED_VARS:
            post = getattr(inf.models, vartype)(name="post_" + latent_var.name)
        else:
            post = inf.models.Deterministic()

        post.base_object = ed_post

        return post


    def __enter__(self):
        """ Method for allowing the use of the construct 'with' """
        ProbModel.__active_models.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Method for allowing the use of the construct 'with' """
        ProbModel.__active_models.pop()

    def add_var(self, v):
        """ Method for adding a new random variable. After use, the model should be re-compiled """

        if not ProbModel.compatible_var(v):
            raise ValueError("The input argument must be a non-generic random variable")

        if v not in self.varlist:
            self.varlist.append(v)
            self.reset_compilation()
    @staticmethod
    def compatible_var(v):
        return (isinstance(v, inf.models.RandomVariable)
                #or isinstance(v, ed.models.RandomVariable)
                ) and not v.is_generic_variable()


    def get_var(self,name):

        """ Get a varible in the model with a given name """

        for v in self.varlist:
            if v.name ==  name:
                return v
        return None



    @tf_run_wrapper
    def log_prob(self, sample_dict):

        """ Computes the log probabilities of a (set of) sample(s)"""

        sd = {}

        for k, v in iteritems(sample_dict):
            var=self.get_var(k)
            sd.update({k: var.log_prob(v, tf_run=False)})


        return sd

    @tf_run_wrapper
    def sum_log_prob(self, sample_dict):

        """ Computes the sum of the log probabilities of a (set of) sample(s)"""

        lp=0

        for k, v in iteritems(sample_dict):
            var = self.get_var(k)
            lp += var.sum_log_prob(v, tf_run=False)

        return lp

    @tf_run_wrapper
    def sample(self, size=1):
        """ Generates a sample for eache variable in the model """
        sd = {}


        for v in self.varlist:
            sd.update({v.name: tf.reshape(v.dist, shape=v.shape)})

        return sd

    def reset_compilation(self):

        """ Clear the structues created during the compilation of the model """

        self.Q = None
        self.data = {}
        self.propagated = False

    def is_compiled(self):
        """ Determines if the model has been compiled """
        return self.Q != None

    def get_config(self):
        raise NotImplementedError

    def to_json(self):
        raise NotImplementedError

    def summary(self):
        raise NotImplementedError


    # static methods
    @staticmethod
    def get_active_model():

        """ Return the active model defined with the construct 'with' """

        if ProbModel.is_active():
            return ProbModel.__active_models[-1]
        return []

    @staticmethod
    def is_active():
        """Check if a replicate construct has been initialized

        Returns:
             True if the method is inside a construct ProbModel (of size different to 1).
             Otherwise False is return
        """
        return len(ProbModel.__active_models)>0


    def get_parents(self, v):

        out = []
        d = self.get_vardict_rev()

        for p in v.dist.get_parents():
            if d.get(p) not in out:
                out.append(d.get(p))

        return out


    def no_parents(self):
        return [v for v in self.varlist if len(v.dist.get_parents()) == 0]

    def get_vardict(self):
        d = {}
        for v in self.varlist:
            d.update({v : v.dist})

        return d


    def get_vardict_rev(self):
        d = {}
        for v in self.varlist:
            d.update({v.dist : v})

        return d


    def copy(self, swap_dict=None):

        new_vars = {} if swap_dict == None else swap_dict

        while len(new_vars.keys()) < len(self.varlist):
            tocopy = [v for v in self.varlist
                      if (len(self.get_parents(v)) == 0 or self.get_parents(v) <= new_vars.keys())
                      and v not in new_vars.keys()]

            v = tocopy[0]

            copy = getattr(inf.models, type(v).__name__)()

            new_vars_ed = {}
            for (key, value) in iteritems(new_vars):
                new_vars_ed.update({key.dist : value.dist if isinstance(value, inf.models.RandomVariable) else value})

            copy.dist = ed.copy(v.dist, new_vars_ed)
            copy.copied_from = v

            new_vars.update({v: copy})

        return ProbModel(varlist=[v for v in new_vars.values() if isinstance(v, inf.models.RandomVariable)])


    def get_copy_from(self, original_var):
        cplst = [v for v in self.varlist if v.copied_from == original_var]

        if len(cplst) == 0:
            return None
        return cplst[0]



    def predict(self, target, data, reset_tf_vars=False):

        # check learnt

        local_hidden = [z for z in target.get_local_hidden() if z not in data.keys()]
        global_hidden = [h for h in self.latent_vars if h not in local_hidden and h not in data.keys()]
        other_observed = [a for a in self.observed_vars if a not in data.keys() and a != target]


        # add posterior of the latent variables
        for h in global_hidden:
            if h not in data.keys():
                data.update({h: self.posterior(h)})

        data_ed = {}
        for (key, value) in iteritems(data):
            data_ed.update(
                {key.dist if isinstance(key, inf.models.RandomVariable) else key :
                     value.dist if isinstance(value, inf.models.RandomVariable) else value})






        q_target = inf.Qmodel.new_qvar(target, check_observed=False)

        latent_vars_ed = {target.dist : q_target.dist}

        for z in local_hidden:
            qz = inf.Qmodel.new_qvar(z, check_observed=False)
            latent_vars_ed.update({z.dist : qz.dist})

        for a in other_observed:
            qa = inf.Qmodel.new_qvar(a, check_observed=False)
            latent_vars_ed.update({a.dist : qa.dist})



        inference_pred = ed.ReparameterizationKLqp(latent_vars_ed, data=data_ed)
        #inference_pred.run()
        inference_pred.initialize()

        sess = inf.util.Runtime.tf_sess

        if reset_tf_vars:
            tf.global_variables_initializer().run()
        else:
            for t in tf.global_variables():
                if not sess.run(tf.is_variable_initialized(t)):
                    sess.run(tf.variables_initializer([t]))


        for _ in range(inference_pred.n_iter):
            info_dict = inference_pred.update()
            inference_pred.print_progress(info_dict)

        inference_pred.finalize()



#        tf.graph_util.convert_variables_to_constants(inf.util.Runtime.tf_sess, tf.get_default_graph())

        return q_target

    def predict_old(self, target_var, observations):

        ### copy qvars from the model

        # check learnt

        # add posterior of the latent variables
        for h in self.latent_vars:
            if h not in observations.keys():
                observations.update({h: self.posterior(h)})

        ancestors = target_var.dist.get_ancestors()
        ancestors_obs = dict([(obs, observations[obs]) for obs in observations.keys() if obs.dist in ancestors])

        m_pred = self.copy(swap_dict=ancestors_obs)

        non_ancestors_obs = dict([(m_pred.get_copy_from(obs), observations[obs]) for obs in observations.keys() if
                                  obs.dist not in ancestors])

        non_ancestors_obs_ed = {}
        for (key, value) in iteritems(non_ancestors_obs):
            non_ancestors_obs_ed.update(
                {key.dist: value.dist if isinstance(value, inf.models.RandomVariable) else value})

        copy_target = m_pred.get_copy_from(target_var)
        q_target = inf.Qmodel.new_qvar(copy_target, check_observed=False)

        inference_pred = ed.KLqp({copy_target.dist: q_target.dist},
                                 data=non_ancestors_obs_ed)

        copy_target.dist.get_parents()

        inference_pred.run()

        return q_target



    def __inf_with_proposal_vars(self):

        import funcsigs
        return 'proposal_vars' in funcsigs.signature(getattr(ed.inferences, self.infMethod)).parameters

        #import inspect  # not working in 2.7
        #return 'proposal_vars' in inspect.signature(getattr(ed.inferences, self.infMethod)).parameters


import edward as ed
import inferpy as inf
from inferpy.models import Normal, InverseGamma

K, d, N = 5, 10, 200

# model definition
with inf.ProbModel() as m:
    #define the weights
    with inf.replicate(size=K):
        w = Normal(0, 1, dim=d)

    sigma = InverseGamma(1.0,1.0)

    # define the generative model
    with inf.replicate(size=N):
        z = Normal(0, 1, dim=K)
        x = Normal(inf.matmul(z,w),
                   sigma, observed=True, dim=d)

# toy data generation
x_train = Normal(loc=0, scale=1., dim=d).sample(N)
data = {x.name: x_train}


# compile and fit the model with training data
m.compile()
m.fit(data)

#extract the hidden representation from a set of observations
hidden_encoding = m.posterior(z)




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Module with the required functionality for implementing the Q-models and Q-distributions used by
some inference algorithms.
"""


import inferpy as inf
import tensorflow as tf
import edward as ed
import sys




class Qmodel(object):

    """Class implementing a Q model

    A Q model approximates the posterior distribution of a probabilistic model P .
    In the ‘Q’ model we should include a q distribution for every non observed variable in the ‘P’ model.
    Otherwise, an error will be raised during model compilation.

    An example of use:

    .. literalinclude:: ../../examples/q_model_inference.py


    """



    def __init__(self, varlist):

        self.varlist = varlist

    """Initializes a Q model

    Args:
        varlist: list with the variables in the model

    """

    @staticmethod
    def build_from_pmodel(p, empirical=False):
        """Initializes a Q model from a P model.

        Args:
            p: P model of type inferpy.ProbModel.
            empirical: determines if q distributions will be empirical or of the same type than each p distribution.

        """
        return inf.Qmodel([inf.Qmodel.new_qvar(v) if not empirical else inf.Qmodel.new_qvar_empirical(v,1000) for v in p.latent_vars])



    @property
    def dict(self):
        """ Dictionary where the keys and values are the p and q distributions respectively. """


        d = {}
        for v in self.varlist:
            d.update({v.bind.dist : v.dist})
        return d


    @property
    def varlist(self):
        """ list of q variables of Inferpy type"""
        return self.__varlist

    @varlist.setter
    def varlist(self,varlist):
        """ modifies the list of variables in the model"""
        for v in varlist:
            if not Qmodel.compatible_var(v):
                raise ValueError("Non-compatible var "+v.name)

        self.__varlist = varlist


    def add_var(self, v):
        """ Method for adding a new q variable """

        if not Qmodel.compatible_var(v):
            raise ValueError("The input argument must be a q variable")

        if v not in self.varlist:
            self.varlist.append(v)



    @staticmethod
    def compatible_var(v):
        return inf.ProbModel.compatible_var(v) and v.bind != None



    @staticmethod
    def __generate_ed_qvar(v, initializer, vartype, params):

        """ Builds an Edward q-variable for a p-variable.

        Args:
            v: Edward variable to be approximated.
            initializer (str): indicates how the new variable should be initialized. Possible values: "ones" , "zeroes".
            vartype (str): Edward type of the new variable.
            params: lists of strings indicating the paramaters of the new variable.


        Returns:
            Edward variable approximating in input variable


        """

        qparams = {}

        if initializer == "ones":
            init_f = tf.ones
        elif initializer == "zeroes":
            init_f = tf.zeros
        else:
            raise ValueError("Unsupported initializer: "+initializer)



        for p_name in params:
            if p_name not in ["probs"]:

                p_shape = getattr(v, p_name).shape.as_list() if hasattr(v, p_name) else v.shape


                var = tf.Variable(init_f(p_shape), dtype="float32", name="q_"+v.name+p_name)
                inf.util.Runtime.tf_sess.run(tf.variables_initializer([var]))


                var = tf.where(tf.is_nan(var), tf.zeros_like(var), var)

                if p_name in ["scale"]:
                    var = tf.nn.softplus(var)
                elif p_name in ["probs"]:
                    var = tf.clip_by_value(tf.where(tf.is_nan(var), tf.zeros_like(var), var),1e-8, 1)
                elif p_name in ["logits"]:
                    var = tf.where(tf.is_nan(var), tf.zeros_like(var), var)

                qparams.update({p_name: var})


        qvar =  getattr(ed.models, vartype)(name = "q_"+str.replace(v.name, ":", ""),allow_nan_stats=False, **qparams)

        return qvar



    @staticmethod
    def new_qvar(v, initializer='ones', qvar_inf_module=None, qvar_inf_type = None, qvar_ed_type = None, check_observed = True, name="qvar"):

        """ Builds an Inferpy q-variable for a p-variable.

        Args:
            v: Inferpy variable to be approximated.
            initializer (str): indicates how the new variable should be initialized. Possible values: "ones" , "zeroes".
            qvar_inf_module (str): module of the new Inferpy variable.
            qvar_inf_type (str): name of the new Inferpy variable.
            qvar_ed_type (str): full name of the encapsulated Edward variable type.
            check_observed (bool): To check if p-variable is observed.

        Returns:
            Inferpy variable approximating in input variable.


        """


        if not inf.ProbModel.compatible_var(v):
            raise ValueError("Non-compatible variable")
        elif v.observed and check_observed:
            raise ValueError("Variable "+v.name+" cannot be observed")

        ## default values ##

        if qvar_inf_module == None:
            qvar_inf_module = sys.modules[v.__class__.__module__]

        if qvar_inf_type == None:
            qvar_inf_type = type(v).__name__

        if qvar_ed_type == None:
            qvar_ed_type = type(v).__name__



        qv = getattr(qvar_inf_module, qvar_inf_type)()
        qv.dist = Qmodel.__generate_ed_qvar(v.dist, initializer, qvar_ed_type, v.PARAMS)
        qv.bind = v
        return qv




    @staticmethod
    def new_qvar_empirical(v, n_post_samples, initializer='ones', check_observed = True, name="qvar"):

        """ Builds an empirical Inferpy q-variable for a p-variable.

        Args:
            v: Inferpy variable to be approximated.
            n_post_samples: number of posterior samples.
            initializer (str): indicates how the new variable should be initialized. Possible values: "ones" , "zeroes".
            check_observed (bool): To check if p-variable is observed.

        Returns:
            Inferpy variable approximating in input variable. The InferPy type will be Deterministic while
            the encapsulated Edward variable will be of type Empirical.


        """


        if not inf.ProbModel.compatible_var(v):
            raise ValueError("Non-compatible variable")
        elif v.observed and check_observed:
            raise ValueError("Variable "+v.name+" cannot be observed")




        qv = inf.models.Deterministic()


        if initializer == "ones":
            init_f = tf.ones
        elif initializer == "zeroes":
            init_f = tf.zeros
        else:
            raise ValueError("Unsupported initializer: "+initializer)


        dtype = v.base_object.dtype.name
        var = tf.Variable(init_f([n_post_samples] + v.shape, dtype=dtype), dtype=dtype, name="q_"+v.name + "/params")
        qv.base_object = ed.models.Empirical(params=var, name = "q_"+str.replace(v.name, ":", ""))


        qv.bind = v
        return qv



    @staticmethod
    def Empirical(v, n_post_samples=500, initializer='ones'):
        return inf.Qmodel.new_qvar_empirical(v, n_post_samples, initializer)


####





def __add__new_qvar(vartype):

    name = vartype.__name__

    def f(cls,v, initializer='ones'):
        return cls.new_qvar(v, initializer, qvar_inf_type=name)


    f.__doc__ = "Creates a new q-variable of type " + name + "" \
                                                             "" \
                                                             "\n\n        Args:" \
                                                             "\n            v: Inferpy p-variable" \
                                                             "\n            initializer (str): indicates how the new variable " \
                                                             "should be initialized. Possible values: 'ones' , 'zeroes'." \
                                                             "";
    f.__name__ = name

    setattr(Qmodel, f.__name__, classmethod(f))


for vartype in inf.models.RandomVariable.__subclasses__():
    __add__new_qvar(vartype)





# -*- coding: utf-8 -*-
#

""" Package with modules dfining functions, classes and variables which are
useful for the main functionality provided by inferpy """


from inferpy.util.runtime import *
from inferpy.util.error import *
from inferpy.util.wrappers import *
from inferpy.util.ops import *
from inferpy.util.format import *




import inferpy as inf
import edward as ed
import tensorflow as tf
import numpy as np
from six import iteritems


sess = ed.get_session()

flag_y = True
f = 2
# graph



N = 1000 # number of observations
# model definition
with inf.ProbModel() as m:
    # prior (latent variable)
    beta = inf.models.Normal(loc=0, scale=1, name="beta")
    w = inf.models.Normal(loc=0, scale=1, name = "w")
    b = inf.models.Normal(loc=0, scale=1, name="b")


    # observed variable
    with inf.replicate(size=N):
        x = inf.models.Normal(loc=beta, scale=1, observed=True, name="x")
        y = inf.models.Normal(loc = w*x+b, scale=1, observed=True, name="y")



# toy data generation

x_train = inf.models.Normal(loc=10, scale=3, dim=1).sample(N)
y_train = x_train*f + inf.models.Normal(loc=1, scale=0.1, dim=1).sample(N)


data = {x.name : x_train, y.name : y_train}

m.compile()
m.fit(data)

qbeta = m.posterior(beta)
qw = m.posterior(w)
qb = m.posterior(b)





sess = ed.get_session()

sess.run([qw.dist])


#####
x_test = inf.models.Normal(loc=10, scale=3, dim=1).sample(N)
y_test = x_test*f + inf.models.Normal(loc=1, scale=0.1, dim=1).sample(N)

x_pred = m.predict(x, data={y:y_test})

x_pred.loc


inf.evaluate("mean_squared_error", n_samples=1000, data={x_pred:x_test}) #0.35903236

inf.criticism.evaluate("log_lik", n_samples=1000, data={x_pred:x_test}) #-1.0081486


inf.evaluate()


import edward as ed
import inferpy as inf


#### learning a 2 parameters of 1-dim from 2-dim data

N = 50
sampling_mean = [30., 10.]
sess = ed.util.get_session()


with inf.ProbModel() as m:

    theta1 = inf.models.Normal(loc=0., scale=1., dim=1)
    theta2 = inf.models.Normal(loc=0., scale=1., dim=1)


    with inf.replicate(size=N):
        x = inf.models.Normal(loc=[theta1, theta2], scale=1., observed=True)


# define the Qmodel



# define with any type
q_theta1 = inf.Qmodel.Uniform(theta1)
q_theta2 = inf.Qmodel.new_qvar(theta2)


qmodel = inf.Qmodel([q_theta1, q_theta2])

inf.Qmodel.Normal(theta1)

m.compile(Q=qmodel)


x_train = inf.models.Normal(loc=sampling_mean, scale=1.).sample(N)
data = {x.name : x_train}


m.fit(data)

m.posterior(theta1).base_object

m.posterior(theta2)







import inferpy as inf

# define a 2-dimension Normal distribution
with inf.replicate(size=2):
	with inf.replicate(size=3):
		x = inf.models.Normal(loc=0., scale=1., dim=2)


# print its parameters
print(x.loc)
print(x.scale)



# the shape of the distribution is (6,2)
print(x.shape)

# get a sample
sample_x = x.sample([4,10])

# the shape of the sample is (4, 10, 6, 2)
print(sample_x.shape)


x.sample(1, tf_run=False)
x.sample(1)

# probability and log probability of the sample
x.prob(sample_x)

x.log_prob(sample_x)


# get the underlying distribution Edward object
ed_x = x.dist






# import the package for using it

import inferpy as inf
import edward as ed
import numpy as np
from six import iteritems


sess = ed.get_session()


N = 1000     # number of observations

# toy data generation

x_train = inf.models.Normal(loc=10, scale=3, dim=1).sample(N)
y_train = x_train*2 + inf.models.Normal(loc=1, scale=0.1, dim=1).sample(N)


# model definition
with inf.ProbModel() as m:
    # prior (latent variable)
    beta = inf.models.Normal(loc=0, scale=1, name="beta")
    w = inf.models.Normal(loc=0, scale=1, name = "w")
    b = inf.models.Normal(loc=0, scale=1, name="b")
    betaz = inf.models.Normal(loc=0, scale=1, name="beta")



    # observed variable
    with inf.replicate(size=N):

        z = inf.models.Normal(loc=betaz, scale=1, observed=False, name="z")
        x = inf.models.Normal(loc=beta+z, scale=1, observed=True, name="x")
        y = inf.models.Normal(loc = w*x+b+z, scale=1, observed=True, name="y")


data = {x.name : x_train, y.name : y_train}

m.compile()
m.fit(data)


x_test = inf.models.Normal(loc=10, scale=3, dim=1).sample(N)
y_test = x_test*2 + inf.models.Normal(loc=1, scale=0.1, dim=1).sample(N)


# predict
y_pred = m.predict(y, data={x : x_test}).loc


# evaluate the predicted data y=y_pred given that x=x_test
mse = inf.evaluate('mean_squared_error', data={x: x_test, y: y_pred}, output_key=y)




import inferpy as inf
import pandas as pd


data = pd.read_csv("inferpy/datasets/test.csv")
N = len(data)

with inf.ProbModel() as m:

    thetaX = inf.models.Normal(loc=0., scale=1.)
    thetaY = inf.models.Normal(loc=0., scale=1.)


    with inf.replicate(size=N):
        x = inf.models.Normal(loc=thetaX, scale=1., observed=True, name="x")
        y = inf.models.Normal(loc=thetaY, scale=1., observed=True, name="y")


m.compile()
m.fit(data)

m.posterior([thetaX, thetaY])



